import dymos as dm

from aviary.mission.phase_builder_base import PhaseBuilderBase, register
from aviary.mission.initial_guess_builders import InitialGuessState, InitialGuessTime, InitialGuessControl

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variable_meta_data import _MetaData
from aviary.mission.flops_based.phases.phase_utils import add_subsystem_variables_to_phase, get_initial
from aviary.variable_info.variables import Dynamic
from aviary.mission.flops_based.ode.mission_ODE import MissionODE


# TODO: support/handle the following in the base class
# - phase.set_time_options()
#     - currently handled in level 3 interface implementation
# - self.external_subsystems
# - self.meta_data, with cls.default_meta_data customization point
@register
class EnergyPhase(PhaseBuilderBase):
    '''
    A phase builder for a simple energy phase.

    Attributes
    ----------
    name : str ('energy_phase')
        object label

    subsystem_options (None)
        dictionary of parameters to be passed to the subsystem builders

    user_options : AviaryValues (<empty>)
        state/path constraint values and flags

        supported options:
            - num_segments : int (5)
                transcription: number of segments
            - order : int (3)
                transcription: order of the state transcription; the order of the control
                transcription is `order - 1`
            - fix_initial : bool (True)
            - initial_ref : float (1.0, 's')
            - initial_bounds : pair ((0.0, 100.0) 's')
            - duration_ref : float (1.0, 's')
            - duration_bounds : pair ((0.0, 100.0) 's')
            - required_available_climb_rate : float (None)
                minimum avaliable climb rate
            - constrain_final : bool (False)
            - input_initial : bool (False)

    initial_guesses : AviaryValues (<empty>)
        state/path beginning values to be set on the problem

        supported options:
            - times
            - distance
            - altitude
            - velocity
            - velocity_rate
            - mass
            - throttle

    ode_class : type (None)
        advanced: the type of system defining the ODE

    transcription : "Dymos transcription object" (None)
        advanced: an object providing the transcription technique of the
        optimal control problem

    external_subsystems : Sequence["subsystem builder"] (<empty>)
        advanced

    meta_data : dict (<"builtin" meta data>)
        advanced: meta data associated with variables in the Aviary data hierarchy

    default_name : str
        class attribute: derived type customization point; the default value
        for name

    default_ode_class : type
        class attribute: derived type customization point; the default value
        for ode_class used by build_phase

    default_meta_data : dict
        class attribute: derived type customization point; the default value for
        meta_data

    Methods
    -------
    build_phase
    make_default_transcription
    validate_options
    assign_default_options
    '''
    __slots__ = ('external_subsystems', 'meta_data')

    # region : derived type customization points
    _meta_data_ = {}

    _initial_guesses_meta_data_ = {}

    default_name = 'cruise'

    default_ode_class = MissionODE

    default_meta_data = _MetaData
    # endregion : derived type customization points

    def __init__(
        self, name=None, subsystem_options=None, user_options=None, initial_guesses=None,
        ode_class=None, transcription=None, core_subsystems=None,
        external_subsystems=None, meta_data=None
    ):
        super().__init__(
            name=name, core_subsystems=core_subsystems, subsystem_options=subsystem_options, user_options=user_options, initial_guesses=initial_guesses, ode_class=ode_class, transcription=transcription)

        # TODO: support external_subsystems and meta_data in the base class
        if external_subsystems is None:
            external_subsystems = []

        self.external_subsystems = external_subsystems

        if meta_data is None:
            meta_data = self.default_meta_data

        self.meta_data = meta_data

    def build_phase(self, aviary_options: AviaryValues = None):
        '''
        Return a new energy phase for analysis using these constraints.

        If ode_class is None, default_ode_class is used.

        If transcription is None, the return value from calling
        make_default_transcription is used.

        Parameters
        ----------
        aviary_options : AviaryValues (<emtpy>)
            collection of Aircraft/Mission specific options

        Returns
        -------
        dymos.Phase
        '''
        phase: dm.Phase = super().build_phase(aviary_options)

        user_options: AviaryValues = self.user_options

        fix_initial = user_options.get_val('fix_initial')
        constrain_final = user_options.get_val('constrain_final')
        optimize_mach = user_options.get_val('optimize_mach')
        optimize_altitude = user_options.get_val('optimize_altitude')
        input_initial = user_options.get_val('input_initial')
        polynomial_control_order = user_options.get_item('polynomial_control_order')[0]
        use_polynomial_control = user_options.get_val('use_polynomial_control')
        throttle_enforcement = user_options.get_val('throttle_enforcement')
        mach_bounds = user_options.get_item('mach_bounds')
        altitude_bounds = user_options.get_item('altitude_bounds')
        initial_mach = user_options.get_item('initial_mach')[0]
        final_mach = user_options.get_item('final_mach')[0]
        initial_altitude = user_options.get_item('initial_altitude')[0]
        final_altitude = user_options.get_item('final_altitude')[0]
        solve_for_distance = user_options.get_val('solve_for_distance')
        no_descent = user_options.get_val('no_descent')
        no_climb = user_options.get_val('no_climb')
        constraints = user_options.get_val('constraints')

        ##############
        # Add States #
        ##############
        # TODO: critically think about how we should handle fix_initial and input_initial defaults.
        # In keeping with Dymos standards, the default should be False instead of True.
        input_initial_mass = get_initial(input_initial, Dynamic.Mission.MASS)
        fix_initial_mass = get_initial(fix_initial, Dynamic.Mission.MASS, True)

        # Experiment: use a constraint for mass instead of connected initial.
        # This is due to some problems in mpi.
        # This is needed for the cutting edge full subsystem integration.
        # TODO: when a Dymos fix is in and we've verified that full case works with the fix,
        # remove this argument.
        if user_options.get_val('add_initial_mass_constraint'):
            phase.add_constraint('rhs_all.initial_mass_residual', equals=0.0, ref=1e4)
            input_initial_mass = False

        phase.add_state(
            Dynamic.Mission.MASS, fix_initial=fix_initial_mass, fix_final=False,
            lower=0.0, ref=1e4, defect_ref=1e6, units='kg',
            rate_source=Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
            targets=Dynamic.Mission.MASS,
            input_initial=input_initial_mass,
            # solve_segments='forward',
        )

        input_initial_distance = get_initial(input_initial, Dynamic.Mission.DISTANCE)
        fix_initial_distance = get_initial(fix_initial, Dynamic.Mission.DISTANCE, True)
        phase.add_state(
            Dynamic.Mission.DISTANCE, fix_initial=fix_initial_distance, fix_final=False,
            lower=0.0, ref=1e6, defect_ref=1e8, units='m',
            rate_source=Dynamic.Mission.DISTANCE_RATE,
            input_initial=input_initial_distance,
            solve_segments='forward' if solve_for_distance else None,
        )

        phase = add_subsystem_variables_to_phase(
            phase, self.name, self.external_subsystems)

        ################
        # Add Controls #
        ################
        if use_polynomial_control:
            phase.add_polynomial_control(
                Dynamic.Mission.MACH,
                targets=Dynamic.Mission.MACH, units=mach_bounds[1],
                opt=optimize_mach, lower=mach_bounds[0][0], upper=mach_bounds[0][1],
                rate_targets=[Dynamic.Mission.MACH_RATE],
                order=polynomial_control_order, ref=0.5,
            )
        else:
            phase.add_control(
                Dynamic.Mission.MACH,
                targets=Dynamic.Mission.MACH, units=mach_bounds[1],
                opt=optimize_mach, lower=mach_bounds[0][0], upper=mach_bounds[0][1],
                rate_targets=[Dynamic.Mission.MACH_RATE],
                ref=0.5,
            )

        # Add altitude rate as a control
        if use_polynomial_control:
            phase.add_polynomial_control(
                Dynamic.Mission.ALTITUDE,
                targets=Dynamic.Mission.ALTITUDE, units=altitude_bounds[1],
                opt=optimize_altitude, lower=altitude_bounds[0][0], upper=altitude_bounds[0][1],
                rate_targets=[Dynamic.Mission.ALTITUDE_RATE],
                order=polynomial_control_order, ref=altitude_bounds[0][1],
            )
        else:
            phase.add_control(
                Dynamic.Mission.ALTITUDE,
                targets=Dynamic.Mission.ALTITUDE, units=altitude_bounds[1],
                opt=optimize_altitude, lower=altitude_bounds[0][0], upper=altitude_bounds[0][1],
                rate_targets=[Dynamic.Mission.ALTITUDE_RATE],
                ref=altitude_bounds[0][1],
            )

        ##################
        # Add Timeseries #
        ##################
        phase.add_timeseries_output(
            Dynamic.Mission.MACH, output_name=Dynamic.Mission.MACH, units='unitless'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.THRUST_TOTAL,
            output_name=Dynamic.Mission.THRUST_TOTAL, units='lbf'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.DRAG, output_name=Dynamic.Mission.DRAG, units='lbf'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.SPECIFIC_ENERGY_RATE_EXCESS,
            output_name=Dynamic.Mission.SPECIFIC_ENERGY_RATE_EXCESS, units='m/s'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
            output_name=Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, units='lbm/h'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.ALTITUDE_RATE,
            output_name=Dynamic.Mission.ALTITUDE_RATE, units='ft/s'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.THROTTLE,
            output_name=Dynamic.Mission.THROTTLE, units='unitless'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.VELOCITY,
            output_name=Dynamic.Mission.VELOCITY, units='m/s'
        )

        ###################
        # Add Constraints #
        ###################
        if optimize_mach and fix_initial and not Dynamic.Mission.MACH in constraints:
            phase.add_boundary_constraint(
                Dynamic.Mission.MACH, loc='initial', equals=initial_mach,
            )

        if optimize_mach and constrain_final and not Dynamic.Mission.MACH in constraints:
            phase.add_boundary_constraint(
                Dynamic.Mission.MACH, loc='final', equals=final_mach,
            )

        if optimize_altitude and fix_initial and not Dynamic.Mission.ALTITUDE in constraints:
            phase.add_boundary_constraint(
                Dynamic.Mission.ALTITUDE, loc='initial', equals=initial_altitude, units=altitude_bounds[1], ref=1.e4,
            )

        if optimize_altitude and constrain_final and not Dynamic.Mission.ALTITUDE in constraints:
            phase.add_boundary_constraint(
                Dynamic.Mission.ALTITUDE, loc='final', equals=final_altitude, units=altitude_bounds[1], ref=1.e4,
            )

        if no_descent and not Dynamic.Mission.ALTITUDE_RATE in constraints:
            phase.add_path_constraint(Dynamic.Mission.ALTITUDE_RATE, lower=0.0)

        if no_climb and not Dynamic.Mission.ALTITUDE_RATE in constraints:
            phase.add_path_constraint(Dynamic.Mission.ALTITUDE_RATE, upper=0.0)

        required_available_climb_rate, units = user_options.get_item(
            'required_available_climb_rate')

        if required_available_climb_rate is not None and not Dynamic.Mission.ALTITUDE_RATE_MAX in constraints:
            phase.add_path_constraint(
                Dynamic.Mission.ALTITUDE_RATE_MAX,
                lower=required_available_climb_rate, units=units
            )

        if not Dynamic.Mission.THROTTLE in constraints:
            if throttle_enforcement == 'boundary_constraint':
                phase.add_boundary_constraint(
                    Dynamic.Mission.THROTTLE, loc='initial', lower=0.0, upper=1.0, units='unitless',
                )
                phase.add_boundary_constraint(
                    Dynamic.Mission.THROTTLE, loc='final', lower=0.0, upper=1.0, units='unitless',
                )
            elif throttle_enforcement == 'path_constraint':
                phase.add_path_constraint(
                    Dynamic.Mission.THROTTLE, lower=0.0, upper=1.0, units='unitless',
                )

        self._add_user_defined_constraints(phase, constraints)

        return phase

    def make_default_transcription(self):
        '''
        Return a transcription object to be used by default in build_phase.
        '''
        user_options = self.user_options

        num_segments, _ = user_options.get_item('num_segments')
        order, _ = user_options.get_item('order')

        seg_ends, _ = dm.utils.lgl.lgl(num_segments + 1)

        transcription = dm.Radau(
            num_segments=num_segments, order=order, compressed=True,
            segment_ends=seg_ends)

        return transcription

    def _extra_ode_init_kwargs(self):
        """
        Return extra kwargs required for initializing the ODE.
        """
        # TODO: support external_subsystems and meta_data in the base class
        return {
            'external_subsystems': self.external_subsystems,
            'meta_data': self.meta_data,
            'subsystem_options': self.subsystem_options,
            'throttle_enforcement': self.user_options.get_val('throttle_enforcement'),
        }


EnergyPhase._add_meta_data(
    'num_segments', val=5, desc='transcription: number of segments')

EnergyPhase._add_meta_data(
    'order', val=3,
    desc='transcription: order of the state transcription; the order of the control'
    ' transcription is `order - 1`')

EnergyPhase._add_meta_data('polynomial_control_order', val=3)

EnergyPhase._add_meta_data('use_polynomial_control', val=True)

EnergyPhase._add_meta_data('add_initial_mass_constraint', val=False)

EnergyPhase._add_meta_data('fix_initial', val=True)

EnergyPhase._add_meta_data('fix_duration', val=False)

EnergyPhase._add_meta_data('optimize_mach', val=False)

EnergyPhase._add_meta_data('optimize_altitude', val=False)

EnergyPhase._add_meta_data('initial_bounds', val=(0., 100.), units='s')

EnergyPhase._add_meta_data('duration_bounds', val=(0., 100.), units='s')

EnergyPhase._add_meta_data(
    'required_available_climb_rate', val=None, units='m/s',
    desc='minimum avaliable climb rate')

EnergyPhase._add_meta_data(
    'no_climb', val=False, desc='aircraft is not allowed to climb during phase')

EnergyPhase._add_meta_data(
    'no_descent', val=False, desc='aircraft is not allowed to descend during phase')

EnergyPhase._add_meta_data('constrain_final', val=False)

EnergyPhase._add_meta_data('input_initial', val=False)

EnergyPhase._add_meta_data('initial_mach', val=None, units='unitless')

EnergyPhase._add_meta_data('final_mach', val=None, units='unitless')

EnergyPhase._add_meta_data('initial_altitude', val=None, units='ft')

EnergyPhase._add_meta_data('final_altitude', val=None, units='ft')

EnergyPhase._add_meta_data('throttle_enforcement', val=None)

EnergyPhase._add_meta_data('mach_bounds', val=(0., 2.), units='unitless')

EnergyPhase._add_meta_data('altitude_bounds', val=(0., 60.e3), units='ft')

EnergyPhase._add_meta_data('solve_for_distance', val=False)

EnergyPhase._add_meta_data('constraints', val={})

EnergyPhase._add_initial_guess_meta_data(
    InitialGuessTime(),
    desc='initial guess for initial time and duration specified as a tuple')

EnergyPhase._add_initial_guess_meta_data(
    InitialGuessState('distance'),
    desc='initial guess for horizontal distance traveled')

EnergyPhase._add_initial_guess_meta_data(
    InitialGuessState('altitude'),
    desc='initial guess for vertical distances')

EnergyPhase._add_initial_guess_meta_data(
    InitialGuessState('mach'),
    desc='initial guess for speed')

EnergyPhase._add_initial_guess_meta_data(
    InitialGuessState('mass'),
    desc='initial guess for mass')



from aviary.interface.methods_for_level2 import AviaryProblem
import os
import pkg_resources

from aviary.examples.external_subsystems.battery.battery_builder import BatteryBuilder
from aviary.examples.external_subsystems.battery.battery_variables import Aircraft
from aviary.api import default_height_energy_phase_info as phase_info


battery_builder = BatteryBuilder(include_constraints=False)

phase_info['pre_mission']['external_subsystems'] = [battery_builder]
phase_info['climb']['external_subsystems'] = [battery_builder]
phase_info['cruise']['external_subsystems'] = [battery_builder]
phase_info['descent']['external_subsystems'] = [battery_builder]

up_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
input_file = pkg_resources.resource_filename(
    "aviary", "models/test_aircraft/aircraft_for_bench_FwFm.csv")

prob = AviaryProblem()

# Load aircraft and options data from user
# Allow for user overrides here
prob.load_inputs(input_file, phase_info)


# Preprocess inputs
prob.check_and_preprocess_inputs()

prob.add_pre_mission_systems()

prob.add_phases()

prob.add_post_mission_systems()

# Link phases and variables
prob.link_phases()

prob.add_driver("SLSQP")

prob.add_design_variables()

prob.add_objective('mass')
# prob.model.add_objective(
#     f'traj.climb.states:{Mission.Battery.STATE_OF_CHARGE}', index=-1, ref=-1)

prob.setup()

prob.set_initial_guesses()

prob.run_aviary_problem()



import unittest

import numpy as np
from openmdao.utils.assert_utils import assert_near_equal

from aviary.examples.variable_meta_data_extension import ExtendedMetaData
from aviary.examples.variables_extension import Aircraft as ExtendedAircraft
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.functions import get_path
from aviary.variable_info.enums import FlapType, GASPEngineType
from aviary.variable_info.variables import Aircraft, Mission


class TestTypes(unittest.TestCase):

    def test_aircraft(self):
        vals = AviaryValues()

        try:
            # if we allow units=None, this test will fail.
            vals.set_val(Aircraft.Design.PART25_STRUCTURAL_CATEGORY, val=3, units=None)
        except TypeError as err:
            self.assertEqual(
                str(err),
                f"AviaryValues: set_val({Aircraft.Design.PART25_STRUCTURAL_CATEGORY}): unsupported units: None"
            )
        else:
            self.fail('Expecting correct units to pass.')

        try:
            vals.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=5, units='unitless')
        except:
            self.fail('Expecting correct units and type to pass.')

        try:
            vals.set_val(Aircraft.Design.COMPUTE_HTAIL_VOLUME_COEFF,
                         val=True, units='unitless')
        except:
            self.fail('Expecting correct units and type to pass.')

        try:
            vals.set_val(Aircraft.Design.COMPUTE_VTAIL_VOLUME_COEFF,
                         val=True, units='unitless')
        except:
            self.fail('Expecting correct units and type to pass.')

        try:
            vals.set_val(Aircraft.CrewPayload.NUM_PASSENGERS,
                         val='five', units='unitless')
        except TypeError as err:
            self.assertEqual(
                str(err),
                f"{Aircraft.CrewPayload.NUM_PASSENGERS} is of type(s) <class 'int'> but"
                " you have provided a value of type <class 'str'>.")
        else:
            self.fail('Expecting TypeError.')

        try:
            vals.set_val(Aircraft.Design.COMPUTE_HTAIL_VOLUME_COEFF,
                         val='five', units='unitless')
        except TypeError as err:
            self.assertEqual(
                str(err),
                f"{Aircraft.Design.COMPUTE_HTAIL_VOLUME_COEFF} is of type(s) <class"
                " 'bool'> but you have provided a value of type <class 'str'>.")
        else:
            self.fail('Expecting TypeError.')

        try:
            vals.set_val(Aircraft.Design.COMPUTE_VTAIL_VOLUME_COEFF,
                         val='five', units='unitless')
        except TypeError as err:
            self.assertEqual(
                str(err),
                f"{Aircraft.Design.COMPUTE_VTAIL_VOLUME_COEFF} is of type(s) <class"
                " 'bool'> but you have provided a value of type <class 'str'>.")
        else:
            self.fail('Expecting TypeError.')

        try:
            vals.set_val(Aircraft.Engine.TYPE, GASPEngineType.TURBOJET)
            self.assertTrue(vals.get_val(Aircraft.Engine.TYPE)
                            is GASPEngineType.TURBOJET)
        except:
            self.fail('Expecting to be able to set the value of an Enum.')

        try:
            vals.set_val(Aircraft.Engine.TYPE, 'turbojet')
            self.assertTrue(vals.get_val(Aircraft.Engine.TYPE)
                            is GASPEngineType.TURBOJET)
        except:
            self.fail('Expecting to be able to set the value of an Enum from a string.')

        try:
            vals.set_val(Aircraft.Engine.TYPE, 'TURBOJET')
            self.assertTrue(vals.get_val(Aircraft.Engine.TYPE)
                            is GASPEngineType.TURBOJET)
        except:
            self.fail('Expecting to be able to set the value of an Enum from a string.')

        try:
            vals.set_val(Aircraft.Engine.TYPE, 7)
            self.assertTrue(vals.get_val(Aircraft.Engine.TYPE)
                            is GASPEngineType.TURBOJET)
        except:
            self.fail('Expecting to be able to set the value of an Enum from an int.')

        try:
            vals.set_val(Aircraft.Engine.TYPE, FlapType.DOUBLE_SLOTTED)
        except ValueError as err:
            self.assertEqual(str(err),
                             "<FlapType.DOUBLE_SLOTTED: 4> is not a valid GASPEngineType")
        else:
            self.fail("Expecting ValueError.")

        try:
            vals.set_val(Aircraft.Engine.DATA_FILE, np.array([]))
        except IndexError as err:
            self.fail("Expecting to be able to set the value of an empty numpy array.")
        else:
            # NOTE comparing two empty numpy arrays will fail (unlike empty lists), so
            #      arrays are appended to for this test
            self.assertEqual(np.append(vals.get_val(
                Aircraft.Engine.DATA_FILE), 0), np.array([0]))

    def test_mission(self):
        vals = AviaryValues()

        try:
            vals.set_val(Mission.Design.CRUISE_ALTITUDE, val=35000, units='ft')
        except:
            self.fail('Expecting correct units and type to pass.')

        try:
            vals.set_val(Mission.Design.CRUISE_ALTITUDE, val=35000.01, units='ft')
        except:
            self.fail('Expecting correct units and type to pass.')

        try:
            vals.set_val(Mission.Design.CRUISE_ALTITUDE, val=True, units='ft')
        except TypeError as err:
            self.assertEqual(
                str(err),
                f"{Mission.Design.CRUISE_ALTITUDE} is of type(s) [<class 'int'>, <class"
                " 'float'>] but you have provided a value of type <class 'bool'>.")
        else:
            self.fail('Expecting TypeError.')

        try:
            vals.set_val(Mission.Design.CRUISE_ALTITUDE, val='fail', units='ft')
        except TypeError as err:
            self.assertEqual(
                str(err),
                f"{Mission.Design.CRUISE_ALTITUDE} is of type(s) [<class 'int'>, <class"
                " 'float'>] but you have provided a value of type <class 'str'>.")
        else:
            self.fail('Expecting TypeError.')


class TestUnits(unittest.TestCase):

    def test_aircraft(self):
        vals = AviaryValues()

        try:
            vals.set_val(Aircraft.Wing.LOADING, val=20, units='N/cm**2')
        except:
            self.fail('Expecting correct units and type to pass.')

        try:
            vals.set_val(Aircraft.Wing.LOADING, val=20, units='kgf/cm**2')
        except ValueError as err:
            self.assertEqual(
                str(err),
                'The units kgf/cm**2 which you have provided for'
                f' {Aircraft.Wing.LOADING} are invalid.')
        else:
            self.fail('Expecting ValueError.')

        try:
            vals.set_val(Aircraft.Wing.LOADING, val=20, units='inch**2/NM')
        except TypeError as err:
            self.assertEqual(
                str(err),
                f"The base units of {Aircraft.Wing.LOADING} are lbf/ft**2, and you have"
                f" tried to set {Aircraft.Wing.LOADING} with units of inch**2/NM, which"
                " are not compatible.")
        else:
            self.fail('Expecting TypeError.')

        try:
            vals.set_val(Aircraft.Wing.LOADING, val=20, units='unitless')
        except TypeError as err:
            self.assertEqual(
                str(err),
                f"The base units of {Aircraft.Wing.LOADING} are lbf/ft**2, and you have"
                f" tried to set {Aircraft.Wing.LOADING} with units of unitless, which"
                " are not compatible.")
        else:
            self.fail('Expecting TypeError.')

    def test_mission(self):
        vals = AviaryValues()

        try:
            vals.set_val(Mission.Takeoff.FINAL_VELOCITY, val=20, units='ft/min')
        except:
            self.fail('Expecting correct units and type to pass.')

        try:
            vals.set_val(Mission.Takeoff.FINAL_VELOCITY, val=20, units='kg')
        except TypeError as err:
            self.assertEqual(
                str(err),
                f"The base units of {Mission.Takeoff.FINAL_VELOCITY} are m/s, and you"
                f" have tried to set {Mission.Takeoff.FINAL_VELOCITY} with units of kg,"
                " which are not compatible.")
        else:
            self.fail('Expecting TypeError.')

        try:
            vals.set_val(Mission.Takeoff.FINAL_VELOCITY, val=20, units='min/ft')
        except TypeError as err:
            self.assertEqual(
                str(err),
                f"The base units of {Mission.Takeoff.FINAL_VELOCITY} are m/s, and you"
                f" have tried to set {Mission.Takeoff.FINAL_VELOCITY} with units of"
                " min/ft, which are not compatible.")
        else:
            self.fail('Expecting TypeError.')


class TestVariableExtension(unittest.TestCase):
    def test_set_val_metadata_extension(self):
        option_defaults = AviaryValues()

        filename = get_path(
            'models/engines/turbofan_23k_1.deck')
        option_defaults.set_val(ExtendedAircraft.Engine.DATA_FILE, filename)
        option_defaults.set_val(ExtendedAircraft.Engine.NUM_ENGINES, val=[
                                2], units='unitless', meta_data=ExtendedMetaData)
        option_defaults.set_val(ExtendedAircraft.Wing.AERO_CENTER, val=5, units='ft', meta_data=ExtendedMetaData
                                )

        check_val = option_defaults.get_val(
            ExtendedAircraft.Wing.AERO_CENTER, units='inch')

        assert_near_equal(check_val, 60, 1e-9)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.enums import FlapType
from aviary.variable_info.variables import Aircraft, Dynamic


class MetaModelGroup(om.Group):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        flap_type = self.options["aviary_options"].get_val(
            Aircraft.Wing.FLAP_TYPE, units='unitless')

        # VDEL1
        VDEL1_interp = self.add_subsystem(
            "VDEL1_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                "aircraft:*",
            ],
            promotes_outputs=[
                "VDEL1",
            ],
        )

        VDEL1_interp.add_input(
            Aircraft.Wing.FLAP_CHORD_RATIO,
            0.3,
            training_data=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5],
            units="unitless",
            desc="ratio of flap chord to wing chord",
        )

        if flap_type is FlapType.PLAIN or flap_type is FlapType.SPLIT:

            VDEL1_interp.add_output(
                "VDEL1",
                1.0,
                training_data=[0.0, 0.32, 0.66, 1.0, 1.32, 1.70],
                units="unitless",
                desc="sensitivity of flap minimum drag coefficient to flap chord ratio",
            )

        else:

            VDEL1_interp.add_output(
                "VDEL1",
                1.0,
                training_data=[0.0, 0.24, 0.55, 1.00, 1.60, 2.20],
                units="unitless",
                desc="sensitivity of flap minimum drag coefficient to flap chord ratio",
            )

        # VDEL2
        VDEL2_interp = self.add_subsystem(
            "VDEL2_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                "flap_defl_ratio",
            ],
            promotes_outputs=[
                "VDEL2",
            ],
        )

        VDEL2_interp.add_input(
            "flap_defl_ratio",
            0.727273,
            training_data=[0.0, 0.25, 0.5, 0.75, 1.0, 1.5, 2.0, 2.25, 2.5, 2.75, 3.0],
            units="unitless",
            desc="ratio of flap deflection to optimum flap deflection angle",
        )

        VDEL2_interp.add_output(
            "VDEL2",
            0.62455,
            training_data=[
                0.0,
                0.18,
                0.37,
                0.65,
                1.00,
                1.97,
                3.44,
                4.15,
                4.55,
                4.82,
                5.00,
            ],
            units="unitless",
            desc="sensitivity of flap minimum drag coefficient to flap angle",
        )

        # VDEL3

        VDEL3_interp = self.add_subsystem(
            "VDEL3_interp",
            om.MetaModelStructuredComp(method="scipy_slinear", extrapolate=True),
            promotes_inputs=["aircraft:*"],
            promotes_outputs=[
                "VDEL3",
            ],
        )

        VDEL3_interp.add_input(
            Aircraft.Wing.FLAP_SPAN_RATIO,
            0.65,
            training_data=[0.0, 0.2, 0.4, 0.6, 0.7, 0.8, 0.9, 1.0],
            units="unitless",
            desc="BTEOB: trailing edge flap span divided by wing span",
        )

        VDEL3_interp.add_input(
            Aircraft.Wing.TAPER_RATIO,
            0.33,
            training_data=[0.0, 0.33, 1.0],
            units="unitless",
            desc="taper ratio of wing",
        )

        VDEL3_interp.add_output(
            "VDEL3",
            0.765,
            units="unitless",
            desc="sensitivity of flap minimum drag coefficient to partial flap span",
            training_data=np.array(
                [
                    [0.0, 0.0, 0.0],
                    [0.4, 0.28, 0.2],
                    [0.67, 0.52, 0.4],
                    [0.86, 0.72, 0.6],
                    [0.92, 0.81, 0.7],
                    [0.96, 0.88, 0.8],
                    [0.99, 0.95, 0.9],
                    [1.0, 1.0, 1.0],
                ]
            ),
        )

        # [0.,.4,.67,.86,.92,.96,.99,1.0,0.,.28,.52,.72,.81,.88,.95,1.0,0.,.2,.4,.6,.7,.8,.9,1.0]

        # VLAM1
        VLAM1_interp = self.add_subsystem(
            "VLAM1_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                "aircraft:*",
            ],
            promotes_outputs=[
                "VLAM1",
            ],
        )

        VLAM1_interp.add_input(
            Aircraft.Wing.ASPECT_RATIO,
            10.13,
            training_data=[
                0.0,
                0.2,
                0.6,
                1.0,
                1.4,
                2.0,
                2.5,
                3.0,
                3.5,
                4.0,
                4.3,
                5.0,
                7.0,
                9.0,
                10.0,
                11.2,
                12.0,
                20.0,
            ],
            units="unitless",
            desc="aspect ratio",
        )

        VLAM1_interp.add_output(
            "VLAM1",
            0.97217,
            training_data=[
                0.0,
                1.36,
                1.47,
                1.49,
                1.47,
                1.24,
                0.97,
                0.91,
                0.88,
                0.87,
                0.86,
                0.87,
                0.92,
                0.96,
                0.97,
                0.99,
                1.0,
                1.0,
            ],
            units="unitless",
            desc="sensitivity of clean wing maximum lift coefficient to wing aspect ratio",
        )

        # VLAM2
        VLAM2_interp = self.add_subsystem(
            "VLAM2_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED,
            ],
            promotes_outputs=["VLAM2"],
        )

        VLAM2_interp.add_input(
            Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED,
            0.13966,
            training_data=[
                0.0,
                0.04,
                0.06,
                0.07,
                0.08,
                0.10,
                0.11,
                0.12,
                0.14,
                0.15,
                0.16,
                0.18,
                0.20,
                0.22,
                0.24,
                0.28,
            ],
            units="unitless",
            desc="average wing thickness to chord ratio",
        )

        VLAM2_interp.add_output(
            "VLAM2",
            1.09948,
            training_data=[
                0.8,
                0.82,
                0.84,
                0.85,
                0.88,
                1.00,
                1.05,
                1.07,
                1.10,
                1.11,
                1.11,
                1.10,
                1.07,
                1.02,
                0.96,
                0.80,
            ],
            units="unitless",
            desc="sensitivity of clean wing maximum lift coefficient to wing thickness to chord ratio",
        )

        # VLAM3
        VLAM3_interp = self.add_subsystem(
            "VLAM3_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                "aircraft:*",
            ],
            promotes_outputs=["VLAM3"],
        )

        VLAM3_interp.add_input(
            Aircraft.Wing.ASPECT_RATIO,
            10.13,
            training_data=[
                0.0,
                0.2,
                0.6,
                1.0,
                1.4,
                2.0,
                2.5,
                3.0,
                3.5,
                4.0,
                4.3,
                5.0,
                7.0,
                9.0,
                10.0,
                11.2,
                12.0,
                20.0,
            ],
            units="unitless",
            desc="aspect ratio",
        )

        VLAM3_interp.add_output(
            "VLAM3",
            0.97217,
            training_data=[
                0.0,
                0.1,
                0.24,
                0.33,
                0.41,
                0.50,
                0.56,
                0.61,
                0.66,
                0.70,
                0.72,
                0.77,
                0.88,
                0.95,
                0.97,
                0.99,
                1.0,
                1.0,
            ],
            units="unitless",
            desc="sensitivity of flap clean wing maximum lift coefficient to wing aspect ratio",
        )

        # VLAM4
        VLAM4_interp = self.add_subsystem(
            "VLAM4_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED,
            ],
            promotes_outputs=[
                "VLAM4",
            ],
        )

        VLAM4_interp.add_input(
            Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED,
            0.13966,
            training_data=[
                0.0,
                0.04,
                0.06,
                0.07,
                0.08,
                0.10,
                0.11,
                0.12,
                0.14,
                0.15,
                0.16,
                0.18,
                0.20,
                0.22,
                0.24,
                0.28,
            ],
            units="unitless",
            desc="average wing thickness to chord ratio",
        )

        if flap_type is FlapType.PLAIN or flap_type is FlapType.SPLIT:

            VLAM4_interp.add_output(
                "VLAM4",
                1.19742,
                training_data=[
                    1.25,
                    1.17,
                    1.08,
                    1.05,
                    1.02,
                    1.00,
                    1.02,
                    1.05,
                    1.20,
                    1.36,
                    1.60,
                    1.87,
                    2.02,
                    2.12,
                    2.18,
                    2.20,
                ],
                units="unitless",
                desc="sensitivity of flap clean wing maximum lift coefficient slope to wing thickness",
            )

        else:

            VLAM4_interp.add_output(
                "VLAM4",
                1.25725,
                training_data=[
                    0.84,
                    0.86,
                    0.89,
                    0.91,
                    0.94,
                    1.00,
                    1.04,
                    1.10,
                    1.26,
                    1.33,
                    1.39,
                    1.49,
                    1.55,
                    1.58,
                    1.59,
                    1.60,
                ],
                units="unitless",
                desc="sensitivity of flap clean wing maximum lift coefficient slope to wing thickness",
            )

        # VLAM5
        VLAM5_interp = self.add_subsystem(
            "VLAM5_intep",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                "aircraft:*",
            ],
            promotes_outputs=["VLAM5"],
        )

        VLAM5_interp.add_input(
            Aircraft.Wing.FLAP_CHORD_RATIO,
            0.3,
            training_data=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5],
            units="unitless",
            desc="ratio of flap chord to wing chord",
        )

        if flap_type is FlapType.PLAIN or flap_type is FlapType.SPLIT:

            VLAM5_interp.add_output(
                "VLAM5",
                1.0,
                training_data=[0.0, 0.72, 0.94, 1.00, 0.95, 0.73],
                units="unitless",
                desc="sensitivity of flap clean wing maximum lift coefficient to wing flap to chord ratio",
            )

        elif (
            flap_type is FlapType.SINGLE_SLOTTED
            or flap_type is FlapType.DOUBLE_SLOTTED
            or flap_type is FlapType.TRIPLE_SLOTTED
        ):

            VLAM5_interp.add_output(
                "VLAM5",
                1.0,
                training_data=[0.0, 0.575, 0.83, 1.00, 1.065, 1.09],
                units="unitless",
                desc="sensitivity of flap clean wing maximum lift coefficient to wing flap to chord ratio",
            )

        else:

            VLAM5_interp.add_output(
                "VLAM5",
                1.0,
                training_data=[0.0, 0.41, 0.73, 1.00, 1.22, 1.40],
                units="unitless",
                desc="sensitivity of flap clean wing maximum lift coefficient to wing flap to chord ratio",
            )

        # VLAM6
        VLAM6_interp = self.add_subsystem(
            "VLAM6_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                "flap_defl",
            ],
            promotes_outputs=[
                "VLAM6",
            ],
        )

        VLAM6_interp.add_input(
            "flap_defl",
            10.0,
            training_data=[
                0.0,
                5.0,
                10.0,
                15.0,
                20.0,
                25.0,
                30.0,
                35.0,
                38.0,
                40.0,
                42.0,
                44.0,
                50.0,
                55.0,
                60.0,
            ],
            units="deg",
            desc="flap deflection",
        )

        if flap_type is FlapType.PLAIN or flap_type is FlapType.SPLIT:

            VLAM6_interp.add_output(
                "VLAM6",
                0.8,
                training_data=[
                    0.0,
                    0.12,
                    0.23,
                    0.34,
                    0.43,
                    0.53,
                    0.62,
                    0.71,
                    0.76,
                    0.80,
                    0.82,
                    0.86,
                    0.94,
                    0.98,
                    1.0,
                ],
                units="unitless",
                desc="sensitivity of flap clean wing maximum lift coefficient to wing flap deflection",
            )

        elif (
            flap_type is FlapType.SINGLE_SLOTTED
            or flap_type is FlapType.DOUBLE_SLOTTED
            or flap_type is FlapType.TRIPLE_SLOTTED
        ):

            VLAM6_interp.add_output(
                "VLAM6",
                1.0,
                training_data=[
                    0.0,
                    0.22,
                    0.41,
                    0.57,
                    0.71,
                    0.83,
                    0.91,
                    0.975,
                    0.995,
                    1.0,
                    0.997,
                    0.992,
                    0.945,
                    0.85,
                    0.75,
                ],
                units="unitless",
                desc="sensitivity of flap clean wing maximum lift coefficient to wing flap deflection",
            )

        elif (flap_type is FlapType.FOWLER or flap_type is FlapType.DOUBLE_SLOTTED_FOWLER):

            VLAM6_interp.add_output(
                "VLAM6",
                1.11,
                training_data=[
                    0.0,
                    0.25,
                    0.46,
                    0.65,
                    0.80,
                    0.92,
                    1.00,
                    1.07,
                    1.10,
                    1.11,
                    1.10,
                    1.07,
                    0.85,
                    0.56,
                    0.20,
                ],
                units="unitless",
                desc="sensitivity of flap clean wing maximum lift coefficient to wing flap deflection",
            )

        else:
            raise ValueError(flap_type + ' is not a valid flap type')

        # VLAM7
        VLAM7_interp = self.add_subsystem(
            "VLAM7_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                Aircraft.Wing.FLAP_SPAN_RATIO,
            ],
            promotes_outputs=[
                "VLAM7",
            ],
        )

        VLAM7_interp.add_input(
            Aircraft.Wing.FLAP_SPAN_RATIO,
            0.65,
            training_data=[0.0, 0.2, 0.4, 0.6, 0.8, 0.9, 1.0],
            units="unitless",
            desc="BTEOB: trailing edge flap span divided by wing span",
        )

        VLAM7_interp.add_output(
            "VLAM7",
            0.735,
            training_data=[0.0, 0.25, 0.47, 0.69, 0.87, 0.94, 1.00],
            units="unitless",
            desc="sensitivity of flap clean wing maximum lift coefficient to wing flap span",
        )

        # VLAM10
        VLAM10_interp = self.add_subsystem(
            "VLAM10_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                "slat_defl_ratio",
            ],
            promotes_outputs=[
                "VLAM10",
            ],
        )

        VLAM10_interp.add_input(
            "slat_defl_ratio",
            0.5,
            training_data=[
                0.0,
                0.2,
                0.4,
                0.5,
                0.6,
                0.7,
                0.8,
                0.9,
                1.0,
                1.1,
                1.2,
                1.4,
                1.6,
                1.7,
            ],
            units="unitless",
            desc="Ratio of leading edge slat deflection to optimum deflection angle",
        )

        VLAM10_interp.add_output(
            "VLAM10",
            0.74,
            training_data=[
                0.0,
                0.34,
                0.62,
                0.74,
                0.83,
                0.90,
                0.96,
                0.99,
                1.00,
                0.99,
                0.96,
                0.81,
                0.49,
                0.22,
            ],
            units="unitless",
            desc="sensitivity of clean wing maximum lift coefficient to slat deflection angle",
        )

        # VLAM11
        VLAM11_interp = self.add_subsystem(
            "VLAM11_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                Aircraft.Wing.SLAT_SPAN_RATIO,
            ],
            promotes_outputs=[
                "VLAM11",
            ],
        )

        VLAM11_interp.add_input(
            Aircraft.Wing.SLAT_SPAN_RATIO,
            0.89759553,
            training_data=[0.0, 0.2, 0.3, 0.4, 0.47, 0.5, 1.0],
            units="unitless",
            desc="ratio of leading edge slat span to wing span",
        )

        VLAM11_interp.add_output(
            "VLAM11",
            0.84232,
            training_data=[0.0, 0.05, 0.09, 0.15, 0.20, 0.23, 1.00],
            units="unitless",
            desc="sensitivity of slat clean wing maximum lift coefficient to slat span",
        )

        # VLAM13
        VLAM13_interp = self.add_subsystem(
            "VLAM13_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                "reynolds",
            ],
            promotes_outputs=[
                "VLAM13",
            ],
        )

        VLAM13_interp.add_input(
            "reynolds",
            val=157.1111,
            training_data=[
                1.0,
                2.0,
                5.0,
                10.0,
                30.0,
                60.0,
                90.0,
                120.0,
                170.0,
                250.0,
                300.0,
                500.0,
                1000.0,
                10000.0,
            ],
            units="unitless",
            desc="reynolds number",
        )

        VLAM13_interp.add_output(
            "VLAM13",
            1.03512,
            training_data=[
                0.70,
                0.70,
                0.75,
                0.81,
                0.925,
                1.0,
                1.04,
                1.05,
                1.03,
                1.00,
                0.98,
                0.93,
                0.90,
                0.90,
            ],
            units="unitless",
            desc="reynolds number correction factor",
        )

        # VLAM14
        VLAM14_interp = self.add_subsystem(
            "VLAM14_interp",
            om.MetaModelStructuredComp(method="1D-slinear", extrapolate=True),
            promotes_inputs=[
                Dynamic.Mission.MACH,
            ],
            promotes_outputs=[
                "VLAM14",
            ],
        )

        VLAM14_interp.add_input(
            Dynamic.Mission.MACH,
            0.17522,
            training_data=[0.0, 0.2, 0.4, 0.6, 0.8, 1.0],
            units="unitless",
            desc="mach number",
        )

        VLAM14_interp.add_output(
            "VLAM14",
            0.99124,
            training_data=[1.0, 0.99, 0.94, 0.87, 0.78, 0.66],
            units="unitless",
            desc="mach number correction factor",
            ref=100,
        )

        # fus_lift
        fus_lift_interp = self.add_subsystem(
            "fus_lift_interp",
            om.MetaModelStructuredComp(method="scipy_slinear", extrapolate=True),
            promotes_inputs=["body_to_span_ratio", "chord_to_body_ratio"],
            promotes_outputs=[
                "fus_lift",
            ],
        )

        fus_lift_interp.add_input(
            "body_to_span_ratio",
            0.09240447,
            training_data=[0.0, 0.05, 0.10, 0.12, 0.15, 0.20, 0.25, 0.30, 0.40, 0.50],
            units="unitless",
            desc="trailing edge flap span divided by wing span",
        )

        fus_lift_interp.add_input(
            "chord_to_body_ratio",
            0.12679,
            training_data=[0.1, 0.2, 0.3, 0.4, 0.5],
            units="unitless",
            desc="taper ratio of wing",
        )

        fus_lift_interp.add_output(
            "fus_lift",
            0.05498,
            units="unitless",
            desc="sensitivity of flap minimum drag coefficient to partial flap span",
            training_data=np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.046, 0.018, -0.002, -0.009, -0.025],
                    [0.070, 0.025, -0.007, -0.030, -0.048],
                    [0.076, 0.026, -0.010, -0.038, -0.057],
                    [0.080, 0.023, -0.018, -0.051, -0.070],
                    [0.073, 0.004, -0.035, -0.073, -0.090],
                    [0.053, -0.022, -0.060, -0.094, -0.109],
                    [0.030, -0.047, -0.084, -0.112, -0.126],
                    [-0.018, -0.094, -0.126, -0.145, -0.155],
                    [-0.068, -0.130, -0.160, -0.172, -0.180],
                ]
            ),
        )



import copy

from aviary.utils.compare_hierarchies import (compare_hierarchies_to_merge,
                                              compare_inner_classes)


def merge_attributes(base_class, merge_class, base_class_attributes, merge_class_attributes):
    '''
    Adds unique attributes of merge_class to base_class.

    For all the attributes of merge_class that are not present in base_class we add them to base_class.
    Attributes present in both classes are ignored because if they are variables they have already
    been checked and found identical and if they are inner classes they will be addressed on a 
    recursive call to this function. Attributes present only in base_class are ignored because we are
    adding to base_class and thus all of its features are automatically preserved.

    Parameters
    ----------
    base_class : class
        A class that is all or part of a variable hierarchy. This can be the top=level
        class in the hierarchy, or any inner class nested at any depth within that top-level class.

    merge_class : class
        A class that is all or part of a variable hierarchy. This can be the top=level
        class in the hierarchy, or any inner class nested at any depth within that top-level class.

    base_class_attributes : set of strings
        A set of the name of all attributes (either variables, inner classes, or both) of base_class.

    merge_class_attributes : set of strings
        A set of the name of all attributes (either variables, inner classes, or both) of merge_class.

    Returns
    -------
    base_class : class
        A class that contains the merged together attributes of base_class and merge_class, with 
        the exception that inner class base_class and merge_class share which have diverging attributes inside 
        of them are not necessarily included.

    Raises
    ------
    None
        No exceptions raised by this method, although other methods called within may raise exceptions.
    '''

    merge_class_unique = merge_class_attributes - \
        base_class_attributes  # attributes present only in merge_class

    for attr in merge_class_unique:
        setattr(base_class, attr, getattr(merge_class, attr))

    return base_class


def recursive_merge(overlapping_inners, base_class, merge_class):
    '''
    Recursively compares all inner classes to an infinite depth and identifies mismatched string-named values.

    For all of the inner class names provided in overlapping_inner_classes this function calls compare_inner_classes
    and compares those inner classes recursively until it reaches the full depth to which outer_class_a and
    outer_class_b have any inner classes in common.

    Parameters
    ----------
    overlapping_inner_classes : set of strings
        This is a set of strings where each string is the name of an inner class that outer_class_a
        and outer_class_b have in common.

    outer_class_a : class
        A class that is all or part of a variable hierarchy. This can be the top-level class in the 
        hierarchy, or any inner class nested at any depth within that top-level class.

    outer_class_b : class
        A class that is all or part of a variable hierarchy. This can be the top-level class in the 
        hierarchy, or any inner class nested at any depth within that top-level class.

    Returns
    ------- 
    None
        No variables returned by this method.

    Raises
    ----------
    None
        No exceptions raised by this method, although other methods called within may raise exceptions.
    '''

    for overlapping_class_name in overlapping_inners:
        overlapping_inner_class_base = getattr(base_class, overlapping_class_name)
        overlapping_inner_class_merge = getattr(merge_class, overlapping_class_name)
        [overlapping_second_inners, vars_base_inner, vars_merge_inner, inners_base_inner, inners_merge_inner] = compare_inner_classes(
            overlapping_inner_class_base, overlapping_inner_class_merge, show_all=True)
        merge_attributes(overlapping_inner_class_base,
                         overlapping_inner_class_merge, vars_base_inner, vars_merge_inner)
        merge_attributes(overlapping_inner_class_base,
                         overlapping_inner_class_merge, inners_base_inner, inners_merge_inner)
        recursive_merge(overlapping_second_inners,
                        overlapping_inner_class_base, overlapping_inner_class_merge)


def merge_two_hierarchies(base_hierarchy, hierarchy_b):
    '''
    Merge two variable hierarchies together by adding the second into the first.

    Add the attributes (variables and inner classes) of two variable hierarchies together, so that
    the attributes that are unique to each hierarchy are combined in the resultant hierarchy. This
    is accomplished by adding on to the first of the two provided hierarchies, and returning it.

    Parameters
    ----------
    base_hierarchy : class
        An Aviary variable hierarchy. This hierarchy will function as the 'base' hierarchy,
        which is the hierarchy that has attributes added onto it from another hierarchy in order to combine
        the attributes of the base and the other hierarchy.

    hierarchy_b : class
        An Aviary variable hierarchy. This hierarchy will function as the 'auxiliary' hierarhcy,
        which is the hierarchy whose unique attributes are added onto a base in order to combine the 
        attributes of the base and the auxiliary.

    Returns
    -------
    base_hierarchy : class
        An Aviary variable hierarchy which includes both the attributes of the inputted base_hierarchy
        and hierarchy_b. This is the same object as the inputted base_hierarchy and has been updated through 
        mutability.

    Raises
    ----------
    None
        No exceptions raised by this method, although other methods called within may raise exceptions.
    '''
    [overlapping_inners, merged_vars, b_vars, merged_inners, b_inners] = compare_inner_classes(
        base_hierarchy, hierarchy_b, show_all=True)
    # this adds the variables of hierarchy_b which are not in base_hierarchy to the overall merged hierarchy
    base_hierarchy = merge_attributes(
        base_hierarchy, hierarchy_b, merged_vars, b_vars)
    # this adds the inner classes of hierarchy_b which are not in base_hierarchy to the overall merged hierarchy
    base_hierarchy = merge_attributes(
        base_hierarchy, hierarchy_b, merged_inners, b_inners)
    recursive_merge(overlapping_inners, base_hierarchy, hierarchy_b)

    return base_hierarchy


def merge_hierarchies(hierarchies_to_merge):
    '''
    Combines all provided variable hierarchies into one unified variable hierarchy.

    Performs checks on the user-provided list of variable hierarchies in order to ensure that they are 
    compatible for merge. Ensures that they are not derived from different superclasses, and that they do
    not have conflicting values. Assuming all checks pass, the provided hierarchies are combined into a
    single hierarchy.

    Parameters
    ----------
    hierarchies_to_merge : list of classes
        This is a list of all the variable hierarchies which should be merged into a single hierarchy. This list should not include hierarchies of multiple types (i.e. an av.Mission hierarchy extension should not be mixed with an av.Aircraft hierarchy extension)

    Returns
    -------
    merged_hierarchy : class
        Aviary variable hierarchy that includes all the variables present in the inputted hierarchy list. Duplicates have been removed, and conflicting duplicates are not possible.

    Raises
    ------
    ValueError
        Raises an exception if the list of inputted variable hierarchies includes hierarchies that have been subclassed from different superclasses.
    '''
    compare_hierarchies_to_merge(hierarchies_to_merge)
    subclass_type = None
    subclass_hierarchy = None
    for hierarchy in hierarchies_to_merge:  # check that there are not hierarchies subclassing from different classes that we are attempting to merge
        # checks if the given hierarchy is the first subclass of the hierarchies
        if (len(hierarchy.__mro__) > 2) and (subclass_type == None):
            # gets highest level superclass of the class before "class" itself
            subclass_type = copy.deepcopy(hierarchy.__mro__[-2])
            subclass_hierarchy = copy.deepcopy(hierarchy)
        elif (len(hierarchy.__mro__) > 2):  # checks if the given hierarchy is a subclass
            if hierarchy.__mro__[-2] != subclass_type:
                raise ValueError(
                    f"You have attempted to merge together variable hierarchies that subclass from different superclasses. '{subclass_hierarchy.__qualname__}' is a subclass of '{subclass_type}' and '{hierarchy.__qualname__}' is a subclass of '{hierarchy.__mro__[-2]}'.")

    if subclass_hierarchy != None:  # ensure that we build on the subclassed hierarchy if we have one, that way the super class information is not lost
        merged_hierarchy = subclass_hierarchy
    else:
        merged_hierarchy = copy.deepcopy(hierarchies_to_merge[0])

    for hierarchy in hierarchies_to_merge:
        merged_hierarchy = merge_two_hierarchies(merged_hierarchy, hierarchy)

    return merged_hierarchy



import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.mass.gasp_based.equipment_and_useful_load import \
    EquipAndUsefulLoadMass
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.enums import GASPEngineType
from aviary.variable_info.variables import Aircraft, Mission


# this is the large single aisle 1 V3 test case
class FixedEquipMassTestCase1(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=180, units='unitless')
        options.set_val(Aircraft.LandingGear.FIXED_GEAR,
                        val=False, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "equip",
            EquipAndUsefulLoadMass(aviary_options=options),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS,
            val=[
                928.0,
                0.0736,
                0.112,
                0.14,
                1959.0,
                1.65,
                551.0,
                11192.0,
                5.0,
                3.0,
                50.0,
                7.6,
                12.0,
            ],
            units="unitless",
        )
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.LENGTH, val=129.4, units="ft"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS, val=7511, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3895.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.AREA, val=375.9, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.AREA, val=469.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=[29500.0], units="lbf"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, val=0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(self.prob[Aircraft.Design.FIXED_EQUIPMENT_MASS], 21089, tol)
        assert_near_equal(
            self.prob[Aircraft.Design.FIXED_USEFUL_LOAD], 5176, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 4932

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=8e-12, rtol=1e-12)


class FixedEquipMassTestCase2(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=5, units='unitless')
        options.set_val(Aircraft.LandingGear.FIXED_GEAR,
                        val=False, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "equip",
            EquipAndUsefulLoadMass(aviary_options=options),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS,
            val=[
                928.0,
                0.0736,
                0.112,
                0.14,
                1959.0,
                1.65,
                551.0,
                11192.0,
                5.0,
                3.0,
                50.0,
                7.6,
                12.0,
            ],
            units="unitless",
        )
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.LENGTH, val=129.4, units="ft"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS, val=7511, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3895.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.AREA, val=375.9, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.AREA, val=469.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=29500.0, units="lbf"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, val=0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        # not actual GASP value
        assert_near_equal(self.prob[Aircraft.Design.FIXED_USEFUL_LOAD], 1235.4, tol)
        assert_near_equal(
            self.prob[Aircraft.Design.FIXED_EQUIPMENT_MASS], 23509, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=8e-12, rtol=1e-12)


class FixedEquipMassTestCase3(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Engine.TYPE,
                        val=[GASPEngineType.RECIP_CARB], units='unitless')
        options.set_val(Aircraft.LandingGear.FIXED_GEAR,
                        val=False, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "equip",
            EquipAndUsefulLoadMass(aviary_options=options),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS,
            val=[
                928.0,
                0.0736,
                0.112,
                0.14,
                1959.0,
                1.65,
                551.0,
                11192.0,
                5.0,
                3.0,
                50.0,
                7.6,
                12.0,
            ],
            units="unitless",
        )
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.LENGTH, val=129.4, units="ft"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS, val=7511, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3895.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.AREA, val=375.9, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.AREA, val=469.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=29500.0, units="lbf"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, val=0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        # not actual GASP value
        assert_near_equal(self.prob[Aircraft.Design.FIXED_USEFUL_LOAD], 892.83, tol)
        assert_near_equal(
            self.prob[Aircraft.Design.FIXED_EQUIPMENT_MASS], 23509, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=8e-12, rtol=1e-12)


class FixedEquipMassTestCase4smooth(
    unittest.TestCase
):  # this is the large single aisle 1 V3 test case
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=180, units='unitless')
        options.set_val(Aircraft.LandingGear.FIXED_GEAR,
                        val=False, units='unitless')
        options.set_val(Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "equip",
            EquipAndUsefulLoadMass(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS,
            val=[
                928.0,
                0.0736,
                0.112,
                0.14,
                1959.0,
                1.65,
                551.0,
                11192.0,
                5.0,
                3.0,
                50.0,
                7.6,
                12.0,
            ],
            units="unitless",
        )
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.LENGTH, val=129.4, units="ft"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS, val=7511, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3895.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.AREA, val=375.9, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.AREA, val=469.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=29500.0, units="lbf"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, val=0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4

        assert_near_equal(self.prob[Aircraft.Design.FIXED_EQUIPMENT_MASS], 21089, tol)
        assert_near_equal(
            self.prob[Aircraft.Design.FIXED_USEFUL_LOAD], 5176, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 4932

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=8e-12, rtol=1e-12)


class FixedEquipMassTestCase5smooth(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=5, units='unitless')
        options.set_val(Aircraft.LandingGear.FIXED_GEAR,
                        val=False, units='unitless')
        options.set_val(Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "equip",
            EquipAndUsefulLoadMass(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS,
            val=[
                928.0,
                0.0736,
                0.112,
                0.14,
                1959.0,
                1.65,
                551.0,
                11192.0,
                5.0,
                3.0,
                50.0,
                7.6,
                12.0,
            ],
            units="unitless",
        )
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.LENGTH, val=129.4, units="ft"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS, val=7511, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3895.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.AREA, val=375.9, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.AREA, val=469.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=29500.0, units="lbf"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, val=0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        # not actual GASP value
        assert_near_equal(self.prob[Aircraft.Design.FIXED_USEFUL_LOAD], 1235.4, tol)
        assert_near_equal(
            self.prob[Aircraft.Design.FIXED_EQUIPMENT_MASS], 23509, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=8e-12, rtol=1e-12)


class FixedEquipMassTestCase6smooth(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=5, units='unitless')
        options.set_val(Aircraft.Engine.TYPE,
                        val=[GASPEngineType.RECIP_CARB], units='unitless')
        options.set_val(Aircraft.LandingGear.FIXED_GEAR,
                        val=False, units='unitless')
        options.set_val(Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES,
                        val=True, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "equip",
            EquipAndUsefulLoadMass(
                aviary_options=options
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS,
            val=[
                928.0,
                0.0736,
                0.112,
                0.14,
                1959.0,
                1.65,
                551.0,
                11192.0,
                5.0,
                3.0,
                50.0,
                7.6,
                12.0,
            ],
            units="unitless",
        )
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.LENGTH, val=129.4, units="ft"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS, val=7511, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3895.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.AREA, val=375.9, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.AREA, val=469.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=29500.0, units="lbf"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, val=0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        # not actual GASP value
        assert_near_equal(self.prob[Aircraft.Design.FIXED_USEFUL_LOAD], 892.83, tol)
        assert_near_equal(
            self.prob[Aircraft.Design.FIXED_EQUIPMENT_MASS], 23509, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=8e-12, rtol=1e-12)


class EquipAndUsefulMassGroupTestCase1(
    unittest.TestCase
):  # this is the large single aisle 1 V3 test case
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=180, units='unitless')
        options.set_val(Aircraft.LandingGear.FIXED_GEAR,
                        val=False, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            EquipAndUsefulLoadMass(aviary_options=options),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Design.EQUIPMENT_MASS_COEFFICIENTS,
            val=[
                928.0,
                0.0736,
                0.112,
                0.14,
                1959.0,
                1.65,
                551.0,
                11192.0,
                5.0,
                3.0,
                50.0,
                7.6,
                12.0,
            ],
            units="unitless",
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )

        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.LENGTH, val=129.4, units="ft"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TOTAL_MASS, val=7511, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.Controls.TOTAL_MASS, val=3895.0, units="lbm")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.AREA, val=375.9, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.AREA, val=469.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5, units="psi")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft")

        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=29500.0, units="lbf"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, val=0.6, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Design.FIXED_USEFUL_LOAD], 5176, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 4932
        assert_near_equal(self.prob[Aircraft.Design.FIXED_EQUIPMENT_MASS], 21089, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=8e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



import unittest
import warnings

from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

from aviary.utils.csv_data_file import write_data_file, read_data_file
from aviary.utils.functions import get_path
from aviary.utils.named_values import NamedValues, get_items, get_keys
from aviary.variable_info.variable_meta_data import CoreMetaData, add_meta_data


@use_tempdirs
class TestAviaryCSV(unittest.TestCase):
    def setUp(self):
        self.filename = get_path('utils/test/csv_test.csv')
        # what will get written to the csv
        self.data = NamedValues({'aircraft:wing:span': ([15.24, 118, 90, 171], 'ft'),
                                 'aircraft:crew_and_payload:num_passengers': ([125, 28, 0.355, 44],
                                                                              'unitless'),
                                 'fake_var': ([0.932, 1023.54, 0, -13], 'lbm')})
        self.comments = ['comment 1', 'comment 2# more comment 2', 'inline comment']

        # what the csv should look like after writing
        self.expected_contents = [
            '# comment 1',
            '# comment 2# more comment 2',
            '# inline comment',
            '',
            'aircraft:wing:span (ft), aircraft:crew_and_payload:num_passengers, fake_var (lbm)',
            '                  15.24,                                      125,          0.932',
            '                    118,                                       28,        1023.54',
            '                     90,                                    0.355,              0',
            '                    171,                                       44,            -13']

    def test_write_data_file(self):
        write_data_file('write.csv', self.data, self.comments)
        read_contents = []
        with open('write.csv') as file:
            for line in file:
                read_contents.append(line.strip('\n'))
        if read_contents != self.expected_contents:
            raise ValueError(f'Contents written to csv do not match '
                             'expected values')

    def test_read_data_file(self):
        self._compare_csv_results(*read_data_file(self.filename, save_comments=True))

    def test_read_metadata_csv(self):
        # catch warnings as errors
        warnings.filterwarnings("error")

        try:
            data, comments = read_data_file(self.filename, CoreMetaData)
        except UserWarning:
            # disable warnings as errors behavior for future tests
            warnings.resetwarnings()
            # run read_data_file() without catching warnings as errors so it completes
            data, comments = read_data_file(
                self.filename, CoreMetaData, save_comments=True)
            if 'fake_var' in get_keys(data):
                raise RuntimeError('fake_var should be skipped when reading csv')
            self._compare_csv_results(data, comments)
        else:
            # disable warnings as errors behavior for future tests
            warnings.resetwarnings()
            raise RuntimeError('File should raise warning of skipped header data')

        # add fake_vars to metadata and try again - should work identically to test_read_data_file()
        add_meta_data(key='fake_var', meta_data=CoreMetaData, units='lbm')
        self._compare_csv_results(
            *read_data_file(self.filename, CoreMetaData, save_comments=True))

    def test_aliases_csv(self):
        aliases = {'Real Var': 'Fake Var'}
        data = read_data_file(
            self.filename, aliases=aliases)
        if 'fake_var' in get_keys(data):
            raise RuntimeError("'fake_var' should be converted to 'Real Var'")
        if 'Real Var' not in get_keys(data):
            raise RuntimeError("'Real Var' is not in data read from csv")

    def _compare_csv_results(self, data, comments):
        expected_data = self.data

        if comments != self.comments:
            raise ValueError(f'Comments read from {self.filename.name} do not match '
                             'expected values')
        for item in get_items(data):
            key = item[0]
            val = item[1][0]
            units = item[1][1]
            if key not in expected_data:
                raise ValueError(f'{key} not found in {self.filename.name}')
            expected_units = expected_data.get_item(key)[1]
            assert_near_equal(val, expected_data.get_val(key, expected_units))
            if units != expected_units:
                raise ValueError(f'Units {units} read from {self.filename.name} do not '
                                 f'match expected units of {expected_units}')


if __name__ == "__main__":
    unittest.main()



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class HorizontalTailMass(om.ExplicitComponent):
    '''
    Calculates the mass of the horizontal tail. The methodology is based on the FLOPS weight
    equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.HorizontalTail.AREA, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.TAPER_RATIO, val=0.0)

        add_aviary_input(self, Mission.Design.GROSS_MASS, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.MASS_SCALER, val=1.0)

        add_aviary_output(self, Aircraft.HorizontalTail.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        area = inputs[Aircraft.HorizontalTail.AREA]
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        scaler = inputs[Aircraft.HorizontalTail.MASS_SCALER]
        taper_ratio = inputs[Aircraft.HorizontalTail.TAPER_RATIO]

        outputs[Aircraft.HorizontalTail.MASS] = scaler * 0.530 * \
            area * gross_weight ** 0.20 * \
            (taper_ratio + 0.50) / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        area = inputs[Aircraft.HorizontalTail.AREA]
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        scaler = inputs[Aircraft.HorizontalTail.MASS_SCALER]
        taper_ratio = inputs[Aircraft.HorizontalTail.TAPER_RATIO]

        gross_weight_exp = gross_weight ** 0.20

        J[Aircraft.HorizontalTail.MASS, Aircraft.HorizontalTail.AREA] = scaler * \
            0.530 * gross_weight_exp * (taper_ratio + 0.50) / GRAV_ENGLISH_LBM

        J[Aircraft.HorizontalTail.MASS, Aircraft.HorizontalTail.MASS_SCALER] = \
            0.530 * area * gross_weight_exp * \
            (taper_ratio + 0.50) / GRAV_ENGLISH_LBM

        J[Aircraft.HorizontalTail.MASS, Mission.Design.GROSS_MASS] = \
            scaler * 0.106 * area * gross_weight ** -0.8 * (taper_ratio + 0.50)

        J[Aircraft.HorizontalTail.MASS, Aircraft.HorizontalTail.TAPER_RATIO] = \
            scaler * 0.530 * area * gross_weight_exp / GRAV_ENGLISH_LBM


class AltHorizontalTailMass(om.ExplicitComponent):
    '''
    Calculates the mass of the horizontal tail using the alternate method.
    The methodology is based on the FLOPS weight equations, modified to
    output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.HorizontalTail.AREA, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.MASS_SCALER, val=1.0)

        add_aviary_output(self, Aircraft.HorizontalTail.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        area = inputs[Aircraft.HorizontalTail.AREA]
        scaler = inputs[Aircraft.HorizontalTail.MASS_SCALER]

        outputs[Aircraft.HorizontalTail.MASS] = scaler * \
            5.4 * area / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        area = inputs[Aircraft.HorizontalTail.AREA]
        scaler = inputs[Aircraft.HorizontalTail.MASS_SCALER]

        J[Aircraft.HorizontalTail.MASS, Aircraft.HorizontalTail.AREA] = (
            5.4 * scaler / GRAV_ENGLISH_LBM
        )

        J[Aircraft.HorizontalTail.MASS, Aircraft.HorizontalTail.MASS_SCALER] = (
            5.4 * area / GRAV_ENGLISH_LBM
        )



"""
This test validates passing a drag polar into the solved aero in the mission.
Computed lift and drag should be the same as reading the same polar in from
a file.
"""
import unittest
import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal

from aviary.interface.methods_for_level2 import AviaryProblem

from aviary.subsystems.subsystem_builder_base import SubsystemBuilderBase
from aviary.utils.csv_data_file import read_data_file
from aviary.utils.named_values import NamedValues
from aviary.interface.default_phase_info.height_energy import phase_info
from aviary.variable_info.variables import Aircraft

from copy import deepcopy


# The drag-polar-generating component reads this in, instead of computing the polars.
polar_file = "subsystems/aerodynamics/gasp_based/data/large_single_aisle_1_aero_free_reduced_alpha.txt"

phase_info = deepcopy(phase_info)

phase_info['pre_mission']['include_takeoff'] = False
phase_info['post_mission']['include_landing'] = False
phase_info['cruise']['subsystem_options']['core_aerodynamics']['method'] = 'solved_alpha'
phase_info['cruise']['subsystem_options']['core_aerodynamics']['aero_data'] = polar_file
phase_info.pop('climb')
phase_info.pop('descent')

data = read_data_file(polar_file)
ALTITUDE = data.get_val('Altitude', 'ft')
MACH = data.get_val('Mach', 'unitless')
ALPHA = data.get_val('Angle_of_Attack', 'deg')

shape = (np.unique(ALTITUDE).size, np.unique(MACH).size, np.unique(ALPHA).size)
CL = data.get_val('CL').reshape(shape)
CD = data.get_val('CD').reshape(shape)


class TestSolvedAero(unittest.TestCase):

    def test_solved_aero_pass_polar(self):
        # Test that passing training data provides the same results
        local_phase_info = deepcopy(phase_info)

        prob = AviaryProblem()

        prob.load_inputs(
            "subsystems/aerodynamics/flops_based/test/data/high_wing_single_aisle.csv", local_phase_info)
        prob.add_pre_mission_systems()
        prob.add_phases()
        prob.add_post_mission_systems()

        prob.link_phases()

        prob.setup()

        prob.set_initial_guesses()

        print('about to run')
        prob.run_model()
        print('done')

        CL_base = prob.get_val("traj.cruise.rhs_all.core_aerodynamics.tabular_aero.CL")
        CD_base = prob.get_val("traj.cruise.rhs_all.core_aerodynamics.tabular_aero.CD")

        # Lift and Drag polars passed from external component in static.

        ph_in = deepcopy(phase_info)

        polar_builder = FakeDragPolarBuilder(name="aero", altitude=ALTITUDE, mach=MACH,
                                             alpha=ALPHA)
        aero_data = NamedValues()
        aero_data.set_val('altitude', ALTITUDE, 'ft')
        aero_data.set_val('mach', MACH, 'unitless')
        aero_data.set_val('angle_of_attack', ALPHA, 'deg')

        subsystem_options = {'method': 'solved_alpha',
                             'aero_data': aero_data,
                             'training_data': True}
        ph_in['pre_mission']['external_subsystems'] = [polar_builder]

        ph_in['cruise']['subsystem_options'] = {'core_aerodynamics': subsystem_options}

        prob = AviaryProblem()

        prob.load_inputs(
            "subsystems/aerodynamics/flops_based/test/data/high_wing_single_aisle.csv", ph_in)

        prob.aviary_inputs.set_val(Aircraft.Design.LIFT_POLAR,
                                   np.zeros_like(CL), units='unitless')
        prob.aviary_inputs.set_val(Aircraft.Design.DRAG_POLAR,
                                   np.zeros_like(CD), units='unitless')

        prob.add_pre_mission_systems()
        prob.add_phases()
        prob.add_post_mission_systems()

        prob.link_phases()

        prob.setup()

        prob.set_initial_guesses()

        prob.run_model()

        CL_pass = prob.get_val("traj.cruise.rhs_all.core_aerodynamics.tabular_aero.CL")
        CD_pass = prob.get_val("traj.cruise.rhs_all.core_aerodynamics.tabular_aero.CD")

        assert_near_equal(CL_pass, CL_base, 1e-6)
        assert_near_equal(CD_pass, CD_base, 1e-6)

    def test_solved_aero_pass_polar_unique_abscissa(self):
        # Solved Aero with shortened lists of table abscissa.
        local_phase_info = deepcopy(phase_info)

        prob = AviaryProblem()

        csv_path = "subsystems/aerodynamics/flops_based/test/data/high_wing_single_aisle.csv"
        prob.load_inputs(csv_path, local_phase_info)
        prob.add_pre_mission_systems()
        prob.add_phases()
        prob.add_post_mission_systems()

        prob.link_phases()

        prob.setup()

        prob.set_initial_guesses()

        prob.run_model()

        CL_base = prob.get_val("traj.cruise.rhs_all.core_aerodynamics.tabular_aero.CL")
        CD_base = prob.get_val("traj.cruise.rhs_all.core_aerodynamics.tabular_aero.CD")

        # Lift and Drag polars passed from external component in static.

        ph_in = deepcopy(phase_info)

        alt = np.array([0., 3000., 6000., 9000., 12000., 15000., 18000., 21000.,
                        24000., 27000., 30000., 33000., 36000., 38000., 42000.])
        mach = np.array([0., 0.2, 0.4, 0.5, 0.6, 0.7, 0.75, 0.8, 0.85, 0.9])
        alpha = np.array([-2., 0., 2., 4., 6., 8., 10.])

        polar_builder = FakeDragPolarBuilder(name="aero", altitude=alt, mach=mach,
                                             alpha=alpha)
        aero_data = NamedValues()
        aero_data.set_val('altitude', alt, 'ft')
        aero_data.set_val('mach', mach, 'unitless')
        aero_data.set_val('angle_of_attack', alpha, 'deg')

        subsystem_options = {'method': 'solved_alpha',
                             'aero_data': aero_data,
                             'training_data': True}
        ph_in['pre_mission']['external_subsystems'] = [polar_builder]

        ph_in['cruise']['subsystem_options'] = {'core_aerodynamics': subsystem_options}

        prob = AviaryProblem()

        prob.load_inputs(csv_path, ph_in)

        prob.aviary_inputs.set_val(Aircraft.Design.LIFT_POLAR,
                                   np.zeros_like(CL), units='unitless')
        prob.aviary_inputs.set_val(Aircraft.Design.DRAG_POLAR,
                                   np.zeros_like(CD), units='unitless')

        prob.add_pre_mission_systems()
        prob.add_phases()
        prob.add_post_mission_systems()

        prob.link_phases()

        prob.setup()

        prob.set_initial_guesses()

        prob.run_model()

        CL_pass = prob.get_val("traj.cruise.rhs_all.core_aerodynamics.tabular_aero.CL")
        CD_pass = prob.get_val("traj.cruise.rhs_all.core_aerodynamics.tabular_aero.CD")

        assert_near_equal(CL_pass, CL_base, 1e-6)
        assert_near_equal(CD_pass, CD_base, 1e-6)


class FakeCalcDragPolar(om.ExplicitComponent):
    """
    This component is a stand-in for an externally computed lift/drag table
    calculation. It does nothing but read in the pre-computed table.
    """

    def initialize(self):
        """
        Declare options.
        """
        self.options.declare("altitude", default=None, allow_none=True,
                             desc="List of altitudes in ascending order.")
        self.options.declare("mach", default=None, allow_none=True,
                             desc="List of mach numbers in ascending order.")
        self.options.declare("alpha", default=None, allow_none=True,
                             desc="List of angles of attack in ascending order.")

    def setup(self):
        altitude = self.options['altitude']
        mach = self.options['mach']
        alpha = self.options['alpha']

        self.add_input(Aircraft.Wing.AREA, 1.0, units='ft**2')
        self.add_input(Aircraft.Wing.SPAN, 1.0, units='ft')

        shape = (len(altitude), len(mach), len(alpha))

        self.add_output('drag_table', shape=shape, units='unitless')
        self.add_output('lift_table', shape=shape, units='unitless')

    def compute(self, inputs, outputs):
        """
        This component doesn't do anything, except set the drag and lift
        polars from the file we read in.

        Any real analysis would compute these tables.
        """
        outputs['drag_table'] = CD
        outputs['lift_table'] = CL


class FakeDragPolarBuilder(SubsystemBuilderBase):
    """
    Prototype of a subsystem that overrides an aviary internally computed var.

    Parameters
    ----------
    altitude : list or None
        List of altitudes in ascending order. (Optional)
    mach : list or None
        List of mach numbers in ascending order. (Optional)
    alpha : list or None
        List of angles of attack in ascending order. (Optional)
    """

    def __init__(self, name='aero', altitude=None, mach=None,
                 alpha=None):
        super().__init__(name)
        self.altitude = np.unique(altitude)
        self.mach = np.unique(mach)
        self.alpha = np.unique(alpha)

    def build_pre_mission(self, aviary_inputs):
        """
        Build an OpenMDAO system for the pre-mission computations of the subsystem.

        Returns
        -------
        pre_mission_sys : openmdao.core.Group
            An OpenMDAO group containing all computations that need to happen in
            the pre-mission (formerly statics) part of the Aviary problem. This
            includes sizing, design, and other non-mission parameters.
        """

        group = om.Group()

        calc_drag_polar = FakeCalcDragPolar(altitude=self.altitude, mach=self.mach,
                                            alpha=self.alpha)

        group.add_subsystem("aero", calc_drag_polar,
                            promotes_inputs=["aircraft:*"],
                            promotes_outputs=[('drag_table', Aircraft.Design.DRAG_POLAR),
                                              ('lift_table', Aircraft.Design.LIFT_POLAR)])
        return group


if __name__ == "__main__":
    # unittest.main()
    test = TestSolvedAero()
    test.test_solved_aero_pass_polar_unique_abscissa()



from copy import deepcopy
import unittest

from openmdao.utils.testing_utils import require_pyoptsparse, use_tempdirs

from aviary.interface.methods_for_level2 import AviaryProblem
from aviary.interface.default_phase_info.two_dof import phase_info as two_dof_phase_info
from aviary.interface.default_phase_info.height_energy import phase_info as height_energy_phase_info
from aviary.interface.default_phase_info.solved import phase_info as solved_phase_info
from aviary.models.N3CC.N3CC_data import inputs


class BaseProblemPhaseTestCase(unittest.TestCase):

    def build_and_run_problem(self, input_filename, phase_info, objective_type=None):
        # Build problem
        prob = AviaryProblem()

        prob.load_inputs(input_filename, phase_info)

        prob.check_and_preprocess_inputs()
        prob.add_pre_mission_systems()
        prob.add_phases()
        prob.add_post_mission_systems()
        prob.link_phases()
        prob.add_driver("SLSQP", max_iter=0)
        prob.add_design_variables()
        prob.add_objective(objective_type if objective_type else None)
        prob.setup()
        prob.set_initial_guesses()
        prob.run_aviary_problem("dymos_solution.db", make_plots=False)


@use_tempdirs
class TwoDOFZeroItersTestCase(BaseProblemPhaseTestCase):

    @require_pyoptsparse(optimizer="IPOPT")
    def test_zero_iters_2DOF(self):
        local_phase_info = deepcopy(two_dof_phase_info)
        self.build_and_run_problem('models/test_aircraft/aircraft_for_bench_GwGm.csv',
                                   local_phase_info)


@use_tempdirs
class HEZeroItersTestCase(BaseProblemPhaseTestCase):

    @require_pyoptsparse(optimizer="IPOPT")
    def test_zero_iters_height_energy(self):
        local_phase_info = deepcopy(height_energy_phase_info)
        local_inputs = deepcopy(inputs)
        local_phase_info['pre_mission']['include_takeoff'] = True
        local_phase_info['post_mission']['include_landing'] = True
        local_phase_info['climb']['user_options']['fix_initial'] = False
        local_phase_info['climb']['user_options']['input_initial'] = True
        self.build_and_run_problem(local_inputs, local_phase_info)


@use_tempdirs
class SolvedZeroItersTestCase(BaseProblemPhaseTestCase):

    @require_pyoptsparse(optimizer="IPOPT")
    def test_zero_iters_solved(self):
        # Modify Aviary inputs before running the common operations
        local_phase_info = deepcopy(solved_phase_info)
        self.build_and_run_problem('models/test_aircraft/aircraft_for_bench_GwGm_solved.csv',
                                   local_phase_info, objective_type="hybrid_objective")


if __name__ == "__main__":
    # unittest.main()
    test = HEZeroItersTestCase()
    test.test_zero_iters_height_energy()



from aviary.subsystems.subsystem_builder_base import SubsystemBuilderBase
from aviary.examples.external_subsystems.battery.model.battery_premission import BatteryPreMission
from aviary.examples.external_subsystems.battery.model.battery_mission import BatteryMission
from aviary.examples.external_subsystems.battery.battery_variables import Aircraft, Mission
from aviary.examples.external_subsystems.battery.battery_variable_meta_data import ExtendedMetaData


class BatteryBuilder(SubsystemBuilderBase):
    '''
    Define the builder for a battery subsystem that provides methods to define the battery subsystem's states, design variables, fixed values, initial guesses, and mass names.

    It also provides methods to build OpenMDAO systems for the pre-mission and mission computations of the subsystem, to get the constraints for the subsystem, and to preprocess the inputs for the subsystem.


    Attributes
    ----------
    name : str ('battery_voltage')
        object label

    Methods
    -------
    __init__(self, name='battery_voltage'):
        Initializes the BatteryBuilder object with a given name.
    get_states(self) -> dict:
        Returns a dictionary of the subsystem's states, where the keys are the names of the state variables, and the values are dictionaries that contain the units for the state variable and any additional keyword arguments required by OpenMDAO for the state variable.
    get_linked_variables(self) -> list:
        Returns an empty list since no variable linking is required for the battery subsystem.
    build_pre_mission(self) -> openmdao.core.System:
        Builds an OpenMDAO system for the pre-mission computations of the subsystem.
    build_mission(self, num_nodes, aviary_inputs) -> openmdao.core.System:
        Builds an OpenMDAO system for the mission computations of the subsystem.
    get_constraints(self) -> dict:
        Returns a dictionary of constraints for the battery subsystem, where the keys are the names of the variables to be constrained, and the values are dictionaries that contain the lower and upper bounds for the constraint and any additional keyword arguments accepted by Dymos for the constraint.
    get_design_vars(self) -> dict:
        Returns a dictionary of design variables for the battery subsystem, where the keys are the names of the design variables, and the values are dictionaries that contain the units for the design variable, the lower and upper bounds for the design variable, and any additional keyword arguments required by OpenMDAO for the design variable.
    get_parameters(self) -> dict:
        Returns a dictionary of fixed values for the battery subsystem, where the keys are the names of the fixed values, and the values are dictionaries that contain the fixed value for the variable, the units for the variable, and any additional keyword arguments required by OpenMDAO for the variable.
    get_initial_guesses(self) -> dict:
        Returns a dictionary of initial guesses for the battery subsystem, where the keys are the names of the initial guesses, and the values are dictionaries that contain the initial guess value, the type of variable (state or control), and any additional keyword arguments required by OpenMDAO for the variable.
    get_mass_names(self) -> list:
        Returns a list of names for the battery subsystem mass.
    preprocess_inputs(self, aviary_inputs) -> aviary_inputs:
        preprocesses the inputs for the battery subsystem, setting the values for battery performance based on the battery cell type.
    '''

    def __init__(self, name='battery', include_constraints=True):
        self.include_constraints = include_constraints
        super().__init__(name, meta_data=ExtendedMetaData)

    def get_states(self):
        '''
        Return a dictionary of states for the battery subsystem.

        Returns
        -------
        states : dict
            A dictionary where the keys are the names of the state variables
            and the values are dictionaries with the following keys:

            - 'units': str
                The units for the state variable.
            - any additional keyword arguments required by OpenMDAO for the state
              variable.
        '''
        states_dict = {
            Mission.Battery.STATE_OF_CHARGE: {
                'rate_source': Mission.Battery.STATE_OF_CHARGE_RATE,
                'fix_initial': True,
            },
            Mission.Battery.VOLTAGE_THEVENIN: {
                'units': 'V',
                'rate_source': Mission.Battery.VOLTAGE_THEVENIN_RATE,
                'defect_ref': 1.e5,
                'ref': 1.e5,
            },
        }

        return states_dict

    def get_linked_variables(self):
        '''
        Return the list of linked variables for the battery subsystem; in this case
        it's our two state variables.
        '''
        return [Mission.Battery.VOLTAGE_THEVENIN, Mission.Battery.STATE_OF_CHARGE]

    def build_pre_mission(self, aviary_inputs):
        '''
        Build an OpenMDAO system for the pre-mission computations of the subsystem.

        Returns
        -------
        pre_mission_sys : openmdao.core.System
            An OpenMDAO system containing all computations that need to happen in
            the pre-mission (formerly statics) part of the Aviary problem. This
            includes sizing, design, and other non-mission parameters.
        '''
        return BatteryPreMission()

    def build_mission(self, num_nodes, aviary_inputs):
        '''
        Build an OpenMDAO system for the mission computations of the subsystem.

        Returns
        -------
        mission_sys : openmdao.core.System
            An OpenMDAO system containing all computations that need to happen
            during the mission. This includes time-dependent states that are
            being integrated as well as any other variables that vary during
            the mission.
        '''
        return BatteryMission(num_nodes=num_nodes, aviary_inputs=aviary_inputs)

    def get_constraints(self):
        '''
        Return a dictionary of constraints for the battery subsystem.

        Returns
        -------
        constraints : dict
            A dictionary where the keys are the names of the variables to be constrained
            and the values are dictionaries are any accepted by Dymos for the
            constraint.

        Description
        -----------
        This method returns a dictionary of constraints for the battery subsystem. The dictionary
        contains two constraints:

        - state of charge constraint: A boundary constraint that ensures the state of charge does not fall
          below 0.2 at the final mission point.
        - thevenin voltage constraint: A path constraint that ensures the Thevenin voltage does not
          fall below 0.
        '''
        if self.include_constraints:
            constraints = {
                Mission.Battery.STATE_OF_CHARGE: {
                    'lower': 0.2,
                    'type': 'boundary',
                    'loc': 'final',
                },
                Mission.Battery.VOLTAGE_THEVENIN: {
                    'lower': 0,
                    'type': 'path',
                },
            }
        else:
            constraints = {}

        return constraints

    def get_design_vars(self):
        '''
        Return a dictionary of design variables for the battery subsystem.

        Returns
        -------
        design_vars : dict
            A dictionary where the keys are the names of the design variables
            and the values are dictionaries with the following keys:

            - 'units': str
                The units for the design variable
            - 'lower': float or None
                The lower bound for the design variable
            - 'upper': float or None
                The upper bound for the design variable
            - any additional keyword arguments required by OpenMDAO for the design
              variable
        '''

        DVs = {
            Aircraft.Battery.Cell.DISCHARGE_RATE: {
                'units': 'A',
                'lower': 0.0,
                'upper': 1.0,
            },
        }

        return DVs

    def get_parameters(self):
        '''
        Return a dictionary of fixed values exposed to the phases for the battery subsystem.

        Returns
        -------
        parameter_info : dict
            A dictionary where the keys are the names of the fixed values
            and the values are dictionaries with the following keys:

            - 'value': float or None
                The fixed value for the variable.
            - 'units': str or None
                The units for the variable.
            - any additional keyword arguments required by OpenMDAO for the variable.
        '''

        parameters_dict = {
            Mission.Battery.TEMPERATURE: {'val': 25.0, 'units': 'degC'},
            Mission.Battery.CURRENT: {'val': 3.25, 'units': 'A'}
        }

        return parameters_dict

    def get_initial_guesses(self):
        '''
        Return a dictionary of initial guesses for the battery subsystem.

        Returns
        -------
        initial_guesses : dict
            A dictionary where the keys are the names of the initial guesses
            and the values are dictionaries with any additional keyword
            arguments required by OpenMDAO for the variable.
        '''

        initial_guess_dict = {
            Mission.Battery.STATE_OF_CHARGE: {
                'val': 1.0,
                'type': 'state',
            },
            Mission.Battery.VOLTAGE_THEVENIN: {
                'val': 5.0,
                'units': 'V',
                'type': 'state',
            },
        }

        return initial_guess_dict

    def get_mass_names(self):
        '''
        Return a list of names for the battery subsystem.

        Returns
        -------
        mass_names : list
            A list of names for the battery subsystem.
        '''
        return [Aircraft.Battery.MASS]

    def preprocess_inputs(self, aviary_inputs):
        '''
        Preprocess the inputs for the battery subsystem.

        Description
        -----------
        This method preprocesses the inputs for the battery subsystem.
        In this case, it sets the values battery performance based on the battery cell type.
        '''
        battery_cell_info_18650 = {
            Aircraft.Battery.Cell.DISCHARGE_RATE: [10.5, 'A'],
            Aircraft.Battery.Cell.ENERGY_CAPACITY_MAX: [3.5, 'A*h'],
            Aircraft.Battery.Cell.HEAT_CAPACITY: [1020.0, 'J/(kg*K)'],
            Aircraft.Battery.Cell.MASS: [0.045, 'kg'],
            Aircraft.Battery.Cell.VOLTAGE_LOW: [2.9, 'V'],
            Aircraft.Battery.Cell.VOLUME: [1.125, 'inch**3']
        }

        for key, val in battery_cell_info_18650.items():
            aviary_inputs.set_val(key, val[0], units=val[1], meta_data=ExtendedMetaData)

        return aviary_inputs

    def get_outputs(self):
        '''
        Return a list of output names for the battery subsystem.

        Returns
        -------
        outputs : list
            A list of variable names for the battery subsystem.
        '''

        return [Mission.Battery.VOLTAGE, Mission.Battery.HEAT_OUT, Aircraft.Battery.EFFICIENCY]



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.flops_based.distributed_prop import \
    distributed_engine_count_factor
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class ElectricalMass(om.ExplicitComponent):
    """
    Computes the mass of the electrical subsystems. The methodology is based on
    the FLOPS weight equations, modified to output mass instead of weight.
    """

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuselage.LENGTH, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.MAX_WIDTH, 0.0)
        add_aviary_input(self, Aircraft.Electrical.MASS_SCALER, 1.0)

        add_aviary_output(self, Aircraft.Electrical.MASS, 1.0)

    def setup_partials(self):
        self.declare_partials(of='*', wrt='*')

    def compute(self, inputs, outputs):
        options: AviaryValues = self.options['aviary_options']
        nfuse = options.get_val(Aircraft.Fuselage.NUM_FUSELAGES)
        ncrew = options.get_val(Aircraft.CrewPayload.NUM_FLIGHT_CREW)
        npass = options.get_val(Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')
        length = inputs[Aircraft.Fuselage.LENGTH]
        width = inputs[Aircraft.Fuselage.MAX_WIDTH]
        num_eng = options.get_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES)
        num_engines_factor = distributed_engine_count_factor(num_eng)
        mass_scaler = inputs[Aircraft.Electrical.MASS_SCALER]

        outputs[Aircraft.Electrical.MASS] = (
            92.0 * length**0.4 * width**0.14 * nfuse**0.27 * num_engines_factor**0.69
            * (1.0 + 0.044 * ncrew + 0.0015 * npass) * mass_scaler / GRAV_ENGLISH_LBM)

    def compute_partials(self, inputs, J):
        options: AviaryValues = self.options['aviary_options']
        nfuse = options.get_val(Aircraft.Fuselage.NUM_FUSELAGES)
        ncrew = options.get_val(Aircraft.CrewPayload.NUM_FLIGHT_CREW)
        npass = options.get_val(Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')
        length = inputs[Aircraft.Fuselage.LENGTH]
        width = inputs[Aircraft.Fuselage.MAX_WIDTH]
        num_eng = options.get_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES)
        num_engines_factor = distributed_engine_count_factor(num_eng)
        mass_scaler = inputs[Aircraft.Electrical.MASS_SCALER]

        fact = 92.0 * nfuse**0.27 * (1.0 + 0.044 * ncrew + 0.0015 * npass)
        length_fact = length**0.4
        width_fact = width**0.14
        ecf_fact = num_engines_factor**0.69

        J[Aircraft.Electrical.MASS, Aircraft.Fuselage.LENGTH] = \
            0.4 * length**-0.6 * fact * width_fact * ecf_fact * \
            mass_scaler / GRAV_ENGLISH_LBM

        J[Aircraft.Electrical.MASS, Aircraft.Fuselage.MAX_WIDTH] = \
            0.14 * width**-0.86 * fact * length_fact * \
            ecf_fact * mass_scaler / GRAV_ENGLISH_LBM

        J[Aircraft.Electrical.MASS, Aircraft.Electrical.MASS_SCALER] = \
            fact * length_fact * ecf_fact * width_fact / GRAV_ENGLISH_LBM


class AltElectricalMass(om.ExplicitComponent):
    """
    Computes the mass of the electrical susbsystems using the alternate method.
    """

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Electrical.MASS_SCALER, 1.0)

        add_aviary_output(self, Aircraft.Electrical.MASS, 1.0)

    def setup_partials(self):
        self.declare_partials(of='*', wrt='*')

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        npass = aviary_options.get_val(
            Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')
        mass_scaler = inputs[Aircraft.Electrical.MASS_SCALER]

        outputs[Aircraft.Electrical.MASS] = 16.3 * \
            npass * mass_scaler / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        npass = aviary_options.get_val(
            Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')

        J[Aircraft.Electrical.MASS, Aircraft.Electrical.MASS_SCALER] = \
            16.3 * npass / GRAV_ENGLISH_LBM



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.propulsion.engine_deck import EngineDeck
from aviary.subsystems.propulsion.engine_sizing import SizeEngine
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.functions import get_path
from aviary.variable_info.variables import Aircraft


class EngineSizingTest1(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem(model=om.Group())

    def test_case_multiengine(self):
        filename = 'models/engines/turbofan_28k.deck'
        filename = get_path(filename)

        options = AviaryValues()
        options.set_val(Aircraft.Engine.DATA_FILE, filename)
        options.set_val(Aircraft.Engine.SCALE_PERFORMANCE, True)
        options.set_val(Aircraft.Engine.SCALE_FACTOR, 1.0)
        options.set_val(Aircraft.Engine.GENERATE_FLIGHT_IDLE, True)
        options.set_val(Aircraft.Engine.IGNORE_NEGATIVE_THRUST, False)
        options.set_val(Aircraft.Engine.FLIGHT_IDLE_THRUST_FRACTION, 0.0)
        options.set_val(Aircraft.Engine.FLIGHT_IDLE_MAX_FRACTION, 1.0)
        options.set_val(Aircraft.Engine.FLIGHT_IDLE_MIN_FRACTION, 0.08)
        options.set_val(Aircraft.Engine.GEOPOTENTIAL_ALT, False)

        engine = EngineDeck(name='engine', options=options)
        options.set_val(Aircraft.Engine.REFERENCE_SLS_THRUST, engine.get_val(
            Aircraft.Engine.REFERENCE_SLS_THRUST, 'lbf'), 'lbf')
        # options.set_val(Aircraft.Engine.SCALE_PERFORMANCE, False)
        # engine2 = EngineDeck(name='engine2', options=options)
        # preprocess_propulsion(options, [engine, engine2])

        self.prob.model.add_subsystem('engine', SizeEngine(
            aviary_options=options), promotes=['*'])
        self.prob.setup(force_alloc_complex=True)
        self.prob.set_val(
            Aircraft.Engine.SCALED_SLS_THRUST,
            np.array([15250]),
            units='lbf')

        self.prob.run_model()

        scale_factor = self.prob.get_val(Aircraft.Engine.SCALE_FACTOR)

        expected_scale_factor = np.array([0.52716908])

        assert_near_equal(scale_factor, expected_scale_factor, tolerance=1e-8)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-10)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_GASP, GRAV_ENGLISH_LBM, MU_TAKEOFF
from aviary.variable_info.enums import AnalysisScheme
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic


class RotationEOM(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("num_nodes", types=int)
        self.options.declare("analysis_scheme", types=AnalysisScheme, default=AnalysisScheme.COLLOCATION,
                             desc="The analysis method that will be used to close the trajectory; for example collocation or time integration")

    def setup(self):
        nn = self.options["num_nodes"]
        analysis_scheme = self.options["analysis_scheme"]

        self.add_input(Dynamic.Mission.MASS, val=np.ones(nn),
                       desc="aircraft mass", units="lbm")
        self.add_input(Dynamic.Mission.THRUST_TOTAL, val=np.ones(
            nn), desc=Dynamic.Mission.THRUST_TOTAL, units="lbf")
        self.add_input(Dynamic.Mission.LIFT, val=np.ones(
            nn), desc=Dynamic.Mission.LIFT, units="lbf")
        self.add_input(Dynamic.Mission.DRAG, val=np.ones(
            nn), desc=Dynamic.Mission.DRAG, units="lbf")
        self.add_input(Dynamic.Mission.VELOCITY, val=np.ones(nn),
                       desc="true air speed", units="ft/s")
        self.add_input(Dynamic.Mission.FLIGHT_PATH_ANGLE, val=np.ones(nn),
                       desc="flight path angle", units="rad")

        add_aviary_input(self, Aircraft.Wing.INCIDENCE, val=0.0, units="deg")
        self.add_input("alpha", val=np.ones(nn), desc="angle of attack", units="deg")

        self.add_output(Dynamic.Mission.VELOCITY_RATE, val=np.ones(nn),
                        desc="TAS rate", units="ft/s**2")
        self.add_output(
            Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE, val=np.ones(nn), desc="flight path angle rate", units="rad/s"
        )
        self.add_output(Dynamic.Mission.ALTITUDE_RATE, val=np.ones(nn),
                        desc="altitude rate", units="ft/s")
        self.add_output(
            Dynamic.Mission.DISTANCE_RATE, val=np.ones(nn), desc="distance rate", units="ft/s"
        )
        self.add_output(
            "normal_force", val=np.ones(nn), desc="normal forces", units="lbf"
        )
        self.add_output(
            "fuselage_pitch", val=np.ones(nn), desc="fuselage pitch angle", units="deg"
        )

        if analysis_scheme is AnalysisScheme.COLLOCATION:
            self.add_output(
                "alpha_rate", val=np.ones(nn), desc="angle of attack rate", units="deg/s"
            )

            self.declare_partials("alpha_rate", ["*"], val=0)
        elif analysis_scheme is AnalysisScheme.SHOOTING:
            self.add_input(
                Dynamic.Mission.DISTANCE, val=np.ones(nn), desc="distance traveled", units="ft"
            )

    def setup_partials(self):
        arange = np.arange(self.options["num_nodes"])

        self.declare_partials(Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE, "*")
        self.declare_partials(
            Dynamic.Mission.VELOCITY_RATE,
            [Dynamic.Mission.THRUST_TOTAL, "alpha", Dynamic.Mission.DRAG,
                Dynamic.Mission.MASS, Dynamic.Mission.FLIGHT_PATH_ANGLE, Dynamic.Mission.LIFT],
            rows=arange,
            cols=arange,
        )
        self.declare_partials(Dynamic.Mission.VELOCITY_RATE, [Aircraft.Wing.INCIDENCE])
        self.declare_partials(Dynamic.Mission.ALTITUDE_RATE, [
                              Dynamic.Mission.VELOCITY, Dynamic.Mission.FLIGHT_PATH_ANGLE], rows=arange, cols=arange)
        self.declare_partials(
            Dynamic.Mission.DISTANCE_RATE, [
                Dynamic.Mission.VELOCITY, Dynamic.Mission.FLIGHT_PATH_ANGLE], rows=arange, cols=arange
        )

        self.declare_partials(
            "normal_force",
            [Dynamic.Mission.MASS, Dynamic.Mission.LIFT,
                Dynamic.Mission.THRUST_TOTAL, "alpha"],
            rows=arange,
            cols=arange,
        )
        self.declare_partials("normal_force", [Aircraft.Wing.INCIDENCE])
        self.declare_partials(
            "fuselage_pitch", Dynamic.Mission.FLIGHT_PATH_ANGLE, rows=arange, cols=arange, val=180 / np.pi,
        )
        self.declare_partials("fuselage_pitch", "alpha", rows=arange, cols=arange, val=1)
        self.declare_partials("fuselage_pitch", Aircraft.Wing.INCIDENCE, val=-1)

    def compute(self, inputs, outputs):
        analysis_scheme = self.options["analysis_scheme"]

        weight = inputs[Dynamic.Mission.MASS] * GRAV_ENGLISH_LBM
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        incremented_lift = inputs[Dynamic.Mission.LIFT]
        incremented_drag = inputs[Dynamic.Mission.DRAG]
        TAS = inputs[Dynamic.Mission.VELOCITY]
        gamma = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]
        i_wing = inputs[Aircraft.Wing.INCIDENCE]
        alpha = inputs["alpha"]

        mu = MU_TAKEOFF

        nn = self.options["num_nodes"]

        thrust_along_flightpath = thrust * np.cos((alpha - i_wing) * np.pi / 180)
        thrust_across_flightpath = thrust * np.sin((alpha - i_wing) * np.pi / 180)
        normal_force = np.clip(weight - incremented_lift - thrust_across_flightpath,
                               a_min=0., a_max=None)

        outputs[Dynamic.Mission.VELOCITY_RATE] = (
            (
                thrust_along_flightpath
                - incremented_drag
                - weight * np.sin(gamma)
                - mu * normal_force
            )
            * GRAV_ENGLISH_GASP
            / weight
        )
        outputs[Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE] = np.zeros(nn)

        outputs[Dynamic.Mission.ALTITUDE_RATE] = TAS * np.sin(gamma)
        outputs[Dynamic.Mission.DISTANCE_RATE] = TAS * np.cos(gamma)
        outputs["normal_force"] = normal_force

        outputs["fuselage_pitch"] = gamma * 180 / np.pi - i_wing + alpha

        if analysis_scheme is AnalysisScheme.COLLOCATION:
            outputs['alpha_rate'] = np.full(nn, 3.33)

    def compute_partials(self, inputs, J):

        mu = MU_TAKEOFF

        weight = inputs[Dynamic.Mission.MASS] * GRAV_ENGLISH_LBM
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        incremented_lift = inputs[Dynamic.Mission.LIFT]
        incremented_drag = inputs[Dynamic.Mission.DRAG]
        TAS = inputs[Dynamic.Mission.VELOCITY]
        gamma = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]
        i_wing = inputs[Aircraft.Wing.INCIDENCE]
        alpha = inputs["alpha"]

        nn = self.options["num_nodes"]

        thrust_along_flightpath = thrust * np.cos((alpha - i_wing) * np.pi / 180)
        thrust_across_flightpath = thrust * np.sin((alpha - i_wing) * np.pi / 180)

        dTAlF_dThrust = np.cos((alpha - i_wing) * np.pi / 180)
        dTAlF_dAlpha = -thrust * np.sin((alpha - i_wing) * np.pi / 180) * np.pi / 180
        dTAlF_dIwing = thrust * np.sin((alpha - i_wing) * np.pi / 180) * np.pi / 180

        dTAcF_dThrust = np.sin((alpha - i_wing) * np.pi / 180)
        dTAcF_dAlpha = thrust * np.cos((alpha - i_wing) * np.pi / 180) * np.pi / 180
        dTAcF_dIwing = -thrust * np.cos((alpha - i_wing) * np.pi / 180) * np.pi / 180

        normal_force = np.clip(weight - incremented_lift - thrust_across_flightpath,
                               a_min=0., a_max=None)

        dNF_dWeight = np.ones(nn)
        dNF_dWeight[normal_force < 0] = 0

        dNF_dLift = -np.ones(nn)
        dNF_dLift[normal_force < 0] = 0

        dNF_dThrust = -np.ones(nn) * dTAcF_dThrust
        dNF_dThrust[normal_force < 0] = 0

        dNF_dAlpha = -np.ones(nn) * dTAcF_dAlpha
        dNF_dAlpha[normal_force < 0] = 0

        dNF_dIwing = -np.ones(nn) * dTAcF_dIwing
        dNF_dIwing[normal_force < 0] = 0

        J[Dynamic.Mission.VELOCITY_RATE, Dynamic.Mission.THRUST_TOTAL] = (
            (dTAlF_dThrust - mu * dNF_dThrust) * GRAV_ENGLISH_GASP / weight
        )
        J[Dynamic.Mission.VELOCITY_RATE, "alpha"] = (
            (dTAlF_dAlpha - mu * dNF_dAlpha) * GRAV_ENGLISH_GASP / weight
        )
        J[Dynamic.Mission.VELOCITY_RATE, Aircraft.Wing.INCIDENCE] = (
            (dTAlF_dIwing - mu * dNF_dIwing) * GRAV_ENGLISH_GASP / weight
        )
        J[Dynamic.Mission.VELOCITY_RATE, Dynamic.Mission.DRAG] = -GRAV_ENGLISH_GASP / weight
        J[Dynamic.Mission.VELOCITY_RATE, Dynamic.Mission.MASS] = (
            GRAV_ENGLISH_GASP * GRAV_ENGLISH_LBM
            * (
                weight * (-np.sin(gamma) - mu * dNF_dWeight)
                - (
                    thrust_along_flightpath
                    - incremented_drag
                    - weight * np.sin(gamma)
                    - mu * normal_force
                )
            )
            / weight**2
        )
        J[Dynamic.Mission.VELOCITY_RATE, Dynamic.Mission.FLIGHT_PATH_ANGLE] = - \
            np.cos(gamma) * GRAV_ENGLISH_GASP
        J[Dynamic.Mission.VELOCITY_RATE, Dynamic.Mission.LIFT] = GRAV_ENGLISH_GASP * \
            (-mu * dNF_dLift) / weight

        J[Dynamic.Mission.ALTITUDE_RATE, Dynamic.Mission.VELOCITY] = np.sin(gamma)
        J[Dynamic.Mission.ALTITUDE_RATE,
            Dynamic.Mission.FLIGHT_PATH_ANGLE] = TAS * np.cos(gamma)

        J[Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.VELOCITY] = np.cos(gamma)
        J[Dynamic.Mission.DISTANCE_RATE,
            Dynamic.Mission.FLIGHT_PATH_ANGLE] = -TAS * np.sin(gamma)

        J["normal_force", Dynamic.Mission.MASS] = dNF_dWeight * GRAV_ENGLISH_LBM
        J["normal_force", Dynamic.Mission.LIFT] = dNF_dLift
        J["normal_force", Dynamic.Mission.THRUST_TOTAL] = dNF_dThrust
        J["normal_force", "alpha"] = dNF_dAlpha
        J["normal_force", Aircraft.Wing.INCIDENCE] = dNF_dIwing



import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.geometry.gasp_based.empennage import (EmpennageSize,
                                                             TailSize,
                                                             TailVolCoef)
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.variables import Aircraft

tol = 5e-4
partial_tols = {"atol": 1e-8, "rtol": 1e-8}


class TestTailVolCoef(
    unittest.TestCase
):  # note, this component is not used in the large single aisle
    def setUp(self):
        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "hvc",
            TailVolCoef(vertical=False),
            promotes_inputs=[
                "aircraft:*",
            ],
        )
        self.prob.model.add_subsystem(
            "vvc",
            TailVolCoef(vertical=True),
            promotes_inputs=[
                "aircraft:*",
            ],
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, val=0, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.LENGTH, val=129.4, units="ft"
        )
        self.prob.model.set_input_defaults("hvc.cab_w", val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults("hvc.wing_ref", val=12.615, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, val=0, units="unitless"
        )
        self.prob.model.set_input_defaults("vvc.cab_w", val=13.1, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults("vvc.wing_ref", 117.8054, units="ft")
        self.prob.setup(check=False, force_alloc_complex=True)

    def test_large_sinle_aisle_1_volcoefs(self):
        self.prob.run_model()
        assert_near_equal(
            self.prob["hvc.vol_coef"], 1.52233, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["vvc.vol_coef"], 0.11623, tol
        )  # not actual GASP value

    def test_partials(self):
        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, **partial_tols)


class TestTailComp(
    unittest.TestCase
):  # this is the GASP test case, input and output values based on large single aisle 1 v3 without bug fix
    def setUp(self):
        self.prob = om.Problem()
        self.prob.model.add_subsystem("tail", TailSize(), promotes=["*"])
        self.prob.setup(check=False, force_alloc_complex=True)

        # values for horizontal tail
        self.prob.model.set_input_defaults("vol_coef", val=1.189, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults("r_arm", val=0.2307, units="unitless")
        self.prob.model.set_input_defaults("wing_ref", val=12.615, units="ft")
        self.prob.model.set_input_defaults("ar", val=4.75, units="unitless")
        self.prob.model.set_input_defaults("tr", val=0.352, units="unitless")

    def test_large_sinle_aisle_1_htail(self):
        self.prob.run_model()

        assert_near_equal(self.prob["area"], 375.9, tol)
        assert_near_equal(self.prob["span"], 42.25, tol)
        assert_near_equal(
            self.prob["rchord"], 13.16130387591471, tol
        )  # (potentially not actual GASP value, it is calculated twice in different places)
        assert_near_equal(self.prob["chord"], 9.57573, tol)
        assert_near_equal(self.prob["arm"], 54.7, tol)

    def test_large_sinle_aisle_1_vtail(self):
        # override horizontal tail defaults for vertical tail
        self.prob.set_val("vol_coef", 0.145, units="unitless")
        self.prob.set_val(Aircraft.Wing.AREA, 1370.3, units="ft**2")
        self.prob.set_val("r_arm", 2.362, units="unitless")
        self.prob.set_val("wing_ref", 117.8, units="ft")
        self.prob.set_val("ar", 1.67, units="unitless")
        self.prob.set_val("tr", 0.801, units="unitless")

        self.prob.run_model()

        assert_near_equal(self.prob["area"], 469.3, tol)
        assert_near_equal(self.prob["span"], 28, tol)
        assert_near_equal(self.prob["rchord"], 18.61267549773935, tol)
        assert_near_equal(self.prob["chord"], 16.83022, tol)
        assert_near_equal(
            self.prob["arm"], 49.87526, tol
        )  # note: slightly different from GASP output value, likely truncation differences, this value is from Kenny

    def test_partials(self):
        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, **partial_tols)


class TestEmpennageGroup(
    unittest.TestCase
):  # this is the GASP test case, input and output values based on large single aisle 1 v3 without bug fix
    def setUp(self):
        self.prob = om.Problem()
        self.prob.model.add_subsystem("emp", EmpennageSize(), promotes=["*"])

        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VOLUME_COEFFICIENT, val=1.189, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MOMENT_RATIO, val=0.2307, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AVERAGE_CHORD, val=12.615, units="ft"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.ASPECT_RATIO, val=4.75, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.TAPER_RATIO, val=0.352, units="unitless"
        )

        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.VOLUME_COEFFICIENT, val=0.145, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MOMENT_RATIO, val=2.362, units="unitless")
        self.prob.model.set_input_defaults(Aircraft.Wing.SPAN, val=117.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.ASPECT_RATIO, val=1.67, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.TAPER_RATIO, val=0.801, units="unitless"
        )

    def test_large_sinle_aisle_1_defaults(self):
        self.prob.model.emp.options["aviary_options"] = get_option_defaults()

        self.prob.setup(check=False, force_alloc_complex=True)

        self.prob.run_model()

        assert_near_equal(self.prob[Aircraft.HorizontalTail.AREA], 375.9, tol)
        assert_near_equal(self.prob[Aircraft.HorizontalTail.SPAN], 42.25, tol)
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.ROOT_CHORD], 13.16130387591471, tol
        )  # (potentially not actual GASP value, it is calculated twice in different places)
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.AVERAGE_CHORD], 9.57573, tol
        )
        assert_near_equal(self.prob[Aircraft.HorizontalTail.MOMENT_ARM], 54.7, tol)

        assert_near_equal(self.prob[Aircraft.VerticalTail.AREA], 469.3, tol)
        assert_near_equal(self.prob[Aircraft.VerticalTail.SPAN], 28, tol)
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.ROOT_CHORD], 18.61267549773935, tol
        )
        assert_near_equal(self.prob[Aircraft.VerticalTail.AVERAGE_CHORD], 16.83022, tol)
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.MOMENT_ARM], 49.87526, tol
        )  # note: slightly different from GASP output value, likely numerical diff.s, this value is from Kenny

    def test_large_sinle_aisle_1_calc_volcoefs(self):
        options = get_option_defaults()
        options.set_val(Aircraft.Design.COMPUTE_HTAIL_VOLUME_COEFF,
                        val=True, units='unitless')
        options.set_val(Aircraft.Design.COMPUTE_VTAIL_VOLUME_COEFF,
                        val=True, units='unitless')
        self.prob.model.emp.options["aviary_options"] = options
        self.prob.setup(check=False, force_alloc_complex=True)

        self.prob.set_val(
            Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, 0, units="unitless")
        self.prob.set_val(Aircraft.Fuselage.LENGTH, 129.4, units="ft")
        self.prob.set_val(Aircraft.Fuselage.AVG_DIAMETER, 13.1, units="ft")

        self.prob.run_model()

        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.VOLUME_COEFFICIENT], 1.52233, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.VOLUME_COEFFICIENT], 0.11623, tol
        )  # not actual GASP value


if __name__ == "__main__":
    unittest.main()
